#include "ilpcommon.hh"


extern "C" {
#include "xed-interface.h"
#include "xed-immdis.h"
#include "xed-portability.h"
}

#include "pin.H"
#include <iostream>
#include <sstream>
//using std::cout;
using std::cerr;
using std::endl;
using std::stringstream;


const char *regx862string(UINT32 i, UINT32 len){
    switch(i){
    case RAX: return "rax";
    case EAX: return "eax";
    case AX: if (len==2) return "ax";
    else return "ah";
    case AL: return "al";
    case RBX: return "rbx";
    case EBX: return "ebx";
    case BX: if (len==2) return "bx";
    else return "bh";
    case BL: return "bl";
    case RCX: return "rcx";
    case ECX: return "ecx";
    case CX: if (len==2) return "cx";
    else return "ch";
    case CL: return "cl";
    case RDX: return "rdx";
    case EDX: return "edx";
    case DX: if (len==2) return "dx";
    else return "dh";
    case DL: return "dl";
    case R8: return "r8";
    case R8D: return "r8d";
    case R8W: return "r8w";
    case R8B: return "r8b";
    case R9: return "r9";
    case R9D: return "r9d";
    case R9W: return "r9w";
    case R9B: return "r9b";
    case R10: return "r10";
    case R10D: return "r10d";
    case R10W: return "r10w";
    case R10B: return "r10b";
    case R11: return "r11";
    case R11D: return "r11d";
    case R11W: return "r11w";
    case R11B: return "r11b";
    case R12: return "r12";
    case R12D: return "r12d";
    case R12W: return "r12w";
    case R12B: return "r12b";
    case R13: return "r13";
    case R13D: return "r13d";
    case R13W: return "r13w";
    case R13B: return "r13b";
    case R14: return "r14";
    case R14D: return "r14d";
    case R14W: return "r14w";
    case R14B: return "r14b";
    case R15: return "r15";
    case R15D: return "r15d";
    case R15W: return "r15w";
    case R15B: return "r15b";
    case RBP: return "rbp";
    case EBP: return "ebp";
    case BP: return "bp";
    case BPL: return "bpl";
    case RSI: return "rsi";
    case ESI: return "esi";
    case SI: return "si";
    case SIL: return "sil";
    case RDI: return "rdi";
    case EDI: return "edi";
    case DI: return "di";
    case DIL: return "dil";
    case RSP: return "rsp";
    case ESP: return "esp";
    case SP: return "sp";
    case SPL: return "spl";
    case CS: return "cs";
    case SS: return "ss";
    case DS: return "ds";
    case ES: return "es";
    case FS: return "fs";
    case GS: return "gs";
    case RFLAGS: return "rflags";
    case RIP: return "rip";
    case MM0: return "mm0";
    case MM1: return "mm1";
    case MM2: return "mm2";
    case MM3: return "mm3";
    case MM4: return "mm4";
    case MM5: return "mm5";
    case MM6: return "mm6";
    case MM7: return "mm7";
    case ST0: return "st(0)";
    case ST1: return "st(1)";
    case ST2: return "st(2)";
    case ST3: return "st(3)";
    case ST4: return "st(4)";
    case ST5: return "st(5)";
    case ST6: return "st(6)";
    case ST7: return "st(7)";
    case XMM0: return "xmm0";
    case XMM1: return "xmm1";
    case XMM2: return "xmm2";
    case XMM3: return "xmm3";
    case XMM4: return "xmm4";
    case XMM5: return "xmm5";
    case XMM6: return "xmm6";
    case XMM7: return "xmm7";
    case XMM8: return "xmm8";
    case XMM9: return "xmm9";
    case XMM10: return "xmm10";
    case XMM11: return "xmm11";
    case XMM12: return "xmm12";
    case XMM13: return "xmm13";
    case XMM14: return "xmm14";
    case XMM15: return "xmm15";
    case YMM0: return "ymm0";
    case YMM1: return "ymm1";
    case YMM2: return "ymm2";
    case YMM3: return "ymm3";
    case YMM4: return "ymm4";
    case YMM5: return "ymm5";
    case YMM6: return "ymm6";
    case YMM7: return "ymm7";
    case YMM8: return "ymm8";
    case YMM9: return "ymm9";
    case YMM10: return "ymm10";
    case YMM11: return "ymm11";
    case YMM12: return "ymm12";
    case YMM13: return "ymm13";
    case YMM14: return "ymm14";
    case YMM15: return "ymm15";

    default: return "unknown";
    }
  }


  UINT32 reg2int(REG r, UINT32 *len){
    unsigned int reg;
    reg = (unsigned int)r;
    switch(r){
    case REG_INVALID_: *len=10000; return 10000;
#if defined(TARGET_IA32E)
    case REG_RDI: *len=8; return RDI;//rdi
    case REG_RSI: *len=8; return RSI;//rsi
    case REG_RBP: *len=8; return RBP;//rbp
    case REG_RSP: *len=8; return RSP;//rsp
    case REG_RBX: *len=8; return RBX;//rbx
    case REG_RDX: *len=8; return RDX;//rdx
    case REG_RCX: *len=8; return RCX;//rcx
    case REG_RAX: *len=8; return RAX;//rax
    case REG_R8: *len=8; return R8;//r8
    case REG_R9: *len=8; return R9;//r9
    case REG_R10: *len=8; return R10;//r10
    case REG_R11: *len=8; return R11;//r11
    case REG_R12: *len=8; return R12;//r12
    case REG_R13: *len=8; return R13;//r13
    case REG_R14: *len=8; return R14;//r14
    case REG_R15: *len=8; return R15;//r15
    case REG_SEG_CS: *len=2; return CS;//cs
    case REG_SEG_SS: *len=2; return SS;//ss
    case REG_SEG_DS: *len=2; return DS;//ds
    case REG_SEG_ES: *len=2; return ES;//es
    case REG_SEG_FS: *len=2; return FS;//fs
    case REG_SEG_GS: *len=2; return GS;//gs
    case REG_RFLAGS: *len=8; return RFLAGS;//rflags
    case REG_RIP: *len=8; return RIP;//rip
#else // ILP_TARGET_IA32E (Registers for TARGET_IA32 without Extension)
      /*
    case REG_EDI: *len=8; return RDI;//rdi
    case REG_ESI: *len=8; return RSI;//rsi
    case REG_EBP: *len=8; return RBP;//rbp
    case REG_ESP: *len=8; return RSP;//rsp
    case REG_EBX: *len=8; return RBX;//rbx
    case REG_EDX: *len=8; return RDX;//rdx
    case REG_ECX: *len=8; return RCX;//rcx
    case REG_EAX: *len=8; return RAX;//rax
      */
    case REG_EDI: *len=4; return EDI;//rdi
    case REG_ESI: *len=4; return ESI;//rsi
    case REG_EBP: *len=4; return EBP;//rbp
    case REG_ESP: *len=4; return ESP;//rsp
    case REG_EBX: *len=4; return EBX;//rbx
    case REG_EDX: *len=4; return EDX;//rdx
    case REG_ECX: *len=4; return ECX;//rcx
    case REG_EAX: *len=4; return EAX;//rax
      /*
    case REG_R8: *len=8; return R8;//r8
    case REG_R9: *len=8; return R9;//r9
    case REG_R10: *len=8; return R10;//r10
    case REG_R11: *len=8; return R11;//r11
    case REG_R12: *len=8; return R12;//r12
    case REG_R13: *len=8; return R13;//r13
    case REG_R14: *len=8; return R14;//r14
    case REG_R15: *len=8; return R15;//r15
      */
    case REG_SEG_CS: *len=2; return CS;//cs
    case REG_SEG_SS: *len=2; return SS;//ss
    case REG_SEG_DS: *len=2; return DS;//ds
    case REG_SEG_ES: *len=2; return ES;//es
    case REG_SEG_FS: *len=2; return FS;//fs
    case REG_SEG_GS: *len=2; return GS;//gs
    case REG_EFLAGS: *len=8; return RFLAGS;//rflags
    case REG_EIP: *len=8; return RIP;//rip      
#endif
    case REG_AL: *len=1; return AL;//al
    case REG_AH: *len=1; return AH;//ah
    case REG_AX: *len=2; return AX;//ax
    case REG_CL: *len=1; return CL;//cl
    case REG_CH: *len=1; return CH;//ch
    case REG_CX: *len=2; return CX;//cx
    case REG_DL: *len=1; return DL;//dl
    case REG_DH: *len=1; return DH;//dh
    case REG_DX: *len=2; return DX;//dx
    case REG_BL: *len=1; return BL;//bl
    case REG_BH: *len=1; return BH;//bh
    case REG_BX: *len=2; return BX;//bx
    case REG_BP: *len=2; return BP;//bp
    case REG_SI: *len=2; return SI;//si
    case REG_DI: *len=2; return DI;//di
    case REG_SP: *len=2; return SP;//sp
    case REG_FLAGS: *len=8; return RFLAGS;//rflags
    case REG_IP: *len=8; return RIP;//rip
      /*
    case 52: *len=1000+reg; return 1000+reg;//
    case 53: *len=1000+reg; return 1000+reg;//
      */
#if defined(TARGET_IA32E)
    case REG_EDI: *len=4; return EDI;//edi
    case REG_DIL: *len=1; return DIL;//dil
    case REG_ESI: *len=4; return ESI;//esi
    case REG_SIL: *len=1; return SIL;//sil
    case REG_EBP: *len=4; return EBP;//ebp
    case REG_BPL: *len=1; return BPL;//bpl
    case REG_ESP: *len=4; return ESP;//esp
    case REG_SPL: *len=1; return SPL;//spl
    case REG_EBX: *len=4; return EBX;//ebx
    case REG_EDX: *len=4; return EDX;//edx
    case REG_ECX: *len=4; return ECX;//ecx
    case REG_EAX: *len=4; return EAX;//eax
    case REG_EFLAGS: *len=8; return RFLAGS;//rflags
    case REG_EIP: *len=8; return RIP;//rip
      /*
    case 66: *len=1000+reg; return 1000+reg;//
    case 67: *len=1000+reg; return 1000+reg;//
      */
    case REG_R8B: *len=1; return R8B;//r8b
    case REG_R8W: *len=2; return R8W;//r8w
    case REG_R8D: *len=4; return R8D;//r8d
    case REG_R9B: *len=1; return R9B;//r9b
    case REG_R9W: *len=2; return R9W;//r9w
    case REG_R9D: *len=4; return R9D;//r9d
    case REG_R10B: *len=1; return R10B;//r10b
    case REG_R10W: *len=2; return R10W;//r10w
    case REG_R10D: *len=4; return R10D;//r10d
    case REG_R11B: *len=1; return R11B;//r11b
    case REG_R11W: *len=2; return R11W;//r11w
    case REG_R11D: *len=4; return R11D;//r11d
    case REG_R12B: *len=1; return R12B;//r12b
    case REG_R12W: *len=2; return R12W;//r12w
    case REG_R12D: *len=4; return R12D;//r12d
    case REG_R13B: *len=1; return R13B;//r13b
    case REG_R13W: *len=2; return R13W;//r13w
    case REG_R13D: *len=4; return R13D;//r13d
    case REG_R14B: *len=1; return R14B;//r14b
    case REG_R14W: *len=2; return R14W;//r14w
    case REG_R14D: *len=4; return R14D;//r14d
    case REG_R15B: *len=1; return R15B;//r15b
    case REG_R15W: *len=2; return R15W;//r15w
    case REG_R15D: *len=4; return R15D;//r15d
#endif

    case REG_MM0: *len=16; return MM0;//mm0
    case REG_MM1: *len=16; return MM1;//mm1
    case REG_MM2: *len=16; return MM2;//mm2
    case REG_MM3: *len=16; return MM3;//mm3
    case REG_MM4: *len=16; return MM4;//mm4
    case REG_MM5: *len=16; return MM5;//mm5
    case REG_MM6: *len=16; return MM6;//mm6
    case REG_MM7: *len=16; return MM7;//mm7
      /*
    case 100: *len=1000+reg; return 1000+reg;//
    case 101: *len=1000+reg; return 1000+reg;//
    case 102: *len=1000+reg; return 1000+reg;//
    case 103: *len=1000+reg; return 1000+reg;//
    case 104: *len=1000+reg; return 1000+reg;//
    case 105: *len=1000+reg; return 1000+reg;//
    case 106: *len=1000+reg; return 1000+reg;//
    case 107: *len=1000+reg; return 1000+reg;//
    case 108: *len=1000+reg; return 1000+reg;//
      */

      //#if defined(TARGET_IA32E)
    case REG_XMM0: *len=16; return XMM0;//xmm0
    case REG_XMM1: *len=16; return XMM1;//xmm1
    case REG_XMM2: *len=16; return XMM2;//xmm2
    case REG_XMM3: *len=16; return XMM3;//xmm3
    case REG_XMM4: *len=16; return XMM4;//xmm4
    case REG_XMM5: *len=16; return XMM5;//xmm5
    case REG_XMM6: *len=16; return XMM6;//xmm6
    case REG_XMM7: *len=16; return XMM7;//xmm7
 
#if defined(TARGET_IA32E)
    case REG_XMM8: *len=16; return XMM8;//xmm8
    case REG_XMM9: *len=16; return XMM9;//xmm9
    case REG_XMM10: *len=16; return XMM10;//xmm10
    case REG_XMM11: *len=16; return XMM11;//xmm11
    case REG_XMM12: *len=16; return XMM12;//xmm12
    case REG_XMM13: *len=16; return XMM13;//xmm13
    case REG_XMM14: *len=16; return XMM14;//xmm14
    case REG_XMM15: *len=16; return XMM15;//xmm15


    case REG_YMM0: *len=32; return YMM0;//ymm0
    case REG_YMM1: *len=32; return YMM1;//ymm0
    case REG_YMM2: *len=32; return YMM2;//ymm0
    case REG_YMM3: *len=32; return YMM3;//ymm0
    case REG_YMM4: *len=32; return YMM4;//ymm0
    case REG_YMM5: *len=32; return YMM5;//ymm0
    case REG_YMM6: *len=32; return YMM6;//ymm0
    case REG_YMM7: *len=32; return YMM7;//ymm0
    case REG_YMM8: *len=32; return YMM8;//ymm0
    case REG_YMM9: *len=32; return YMM9;//ymm0
    case REG_YMM10: *len=32; return YMM10;//ymm0
    case REG_YMM11: *len=32; return YMM11;//ymm0
    case REG_YMM12: *len=32; return YMM12;//ymm0
    case REG_YMM13: *len=32; return YMM13;//ymm0
    case REG_YMM14: *len=32; return YMM14;//ymm0
    case REG_YMM15: *len=32; return YMM15;//ymm0


#endif
      /*
    case 125: *len=1000+reg; return 1000+reg;//
    case 126: *len=1000+reg; return 1000+reg;//
    case 127: *len=1000+reg; return 1000+reg;//
    case 128: *len=1000+reg; return 1000+reg;//
    case 129: *len=1000+reg; return 1000+reg;//
    case 130: *len=1000+reg; return 1000+reg;//
    case 131: *len=1000+reg; return 1000+reg;//
    case 132: *len=1000+reg; return 1000+reg;//
    case 133: *len=1000+reg; return 1000+reg;//
    case 134: *len=1000+reg; return 1000+reg;//
    case 135: *len=1000+reg; return 1000+reg;//
    case 136: *len=1000+reg; return 1000+reg;//
    case 137: *len=1000+reg; return 1000+reg;//
    case 138: *len=1000+reg; return 1000+reg;//
    case 139: *len=1000+reg; return 1000+reg;//
    case 140: *len=1000+reg; return 1000+reg;//
    case 141: *len=1000+reg; return 1000+reg;//
    case 142: *len=1000+reg; return 1000+reg;//
    case 143: *len=1000+reg; return 1000+reg;//
    case 144: *len=1000+reg; return 1000+reg;//
    case 145: *len=1000+reg; return 1000+reg;//
    case 146: *len=1000+reg; return 1000+reg;//
    case 147: *len=1000+reg; return 1000+reg;//
    case 148: *len=1000+reg; return 1000+reg;//
    case 149: *len=1000+reg; return 1000+reg;//
    case 150: *len=1000+reg; return 1000+reg;//
    case 151: *len=1000+reg; return 1000+reg;//
    case 152: *len=1000+reg; return 1000+reg;//
    case 153: *len=1000+reg; return 1000+reg;//
    case 154: *len=1000+reg; return 1000+reg;//
      */ 
    case REG_ST0: *len=16; return ST0;//st(0)
    case REG_ST1: *len=16; return ST1;//st(1)
    case REG_ST2: *len=16; return ST2;//st(2)
    case REG_ST3: *len=16; return ST3;//st(3)
    case REG_ST4: *len=16; return ST4;//st(4)
    case REG_ST5: *len=16; return ST5;//st(5)
    case REG_ST6: *len=16; return ST6;//st(6)
    case REG_ST7: *len=16; return ST7;//st(7)

    default: *len=1000+reg; return 1000+reg;//
    }
  }

//using LEVEL_BASE::REG_RDI;
Register::Register(): num(0), len(0) {}
Instruction::Instruction(): address(0L), is_control(false), opcode(0), nbr(0), nbw(0),
			    rr(), wr(),
			    amr(0L), rsize(0), amr2(0L), amw(0L), wsize(0), xed_category(XED_CATEGORY_INVALID)
{
  for (int i=0; i<8; i++)
    {
      rr[i] = Register();
      wr[i] = Register();
    }
}

Instruction::Instruction(ADDRINT addr, bool _is_control, UINT32 op, 
			 uint32_t nr, uint32_t nw, Register rreg[8], Register wreg[8],
			 ADDRINT _amr,
                         UINT32 _rsize, 
			 ADDRINT _amr2,
			 ADDRINT _amw,
                         UINT32 _wsize,
			 xed_category_enum_t _xed_category
			 )
{
  address = addr;
  is_control = _is_control;
  opcode = op;
  nbr = nr;
  nbw = nw;
  amr = _amr;
  rsize = _rsize;
  amr2 = _amr2;
  amw = _amw;
  wsize = _wsize;
  for (int i=0; i<8; i++)
    {
      rr[i] = rreg[i];
      wr[i] = wreg[i];
    }
  xed_category = _xed_category;
}

void Instruction::dump(ostream &os)
{
  os << "[@:" 
     << hex << address << dec << ":"
     << hex << opcode  << dec  << ":"
    // TODO Category
     << xed_category << ":"
     << hex << nbr     << dec  << ":"
     << hex << nbw     << dec  << ":";
  for (unsigned int i=0; i<nbr; i++)
    os << "rr"<<i<<"(" << rr[i].num << "),";
  os << ":";
  for (unsigned int i=0; i<nbw; i++)
    os << "wr"<<i<<"(" << wr[i].num << "),";
  os << ":";
  os << hex << amr   << dec  << ":"
     << hex << amr2  << dec  << ":"
     << hex << amw   << dec  << ":";
  os << "]";

}
