#ifndef ILPCOMMON_HH
#define ILPCOMMON_HH
#include "pin.H"
#include "utils/memory.hh"
#include "utils/memory.tcc"

extern "C" {
#include "xed-category-enum.h"
}

#include "config.hh"

typedef unsigned long ULONG;

#define LOG_PSIZE 18
#define PAGE_SIZE (1<<LOG_PSIZE)
#define PAGE_MASK (PAGE_SIZE-1)

#define REG_FILE_SIZE 2048
//#define MAX_PAGES 1024
#define MAX_PAGES 64
/*
#define MAX_STORES (TRACE_LENGTH/4)
*/
#define WRITE 0
#define READ 1
//#define NBREGS 108
#define NBREGS 124 // 108 + 16 YMM registers

#define RAX 16
#define EAX 20
#define AX 22
#define AH 22
#define AL 23
#define RBX 24
#define EBX 28
#define BX 30
#define BH 30
#define BL 31
#define RCX 32
#define ECX 36
#define CX 38
#define CH 38
#define CL 39
#define RDX 40
#define EDX 44
#define DX 46
#define DH 46
#define DL 47
#define R8 48
#define R8D 52
#define R8W 54
#define R8B 55
#define R9 56
#define R9D 60
#define R9W 62
#define R9B 63
#define R10 64
#define R10D 68
#define R10W 70
#define R10B 71
#define R11 72
#define R11D 76
#define R11W 78
#define R11B 79
#define R12 80
#define R12D 84
#define R12W 86
#define R12B 87
#define R13 88
#define R13D 92
#define R13W 94
#define R13B 95
#define R14 96
#define R14D 100
#define R14W 102
#define R14B 103
#define R15 104
#define R15D 108
#define R15W 110
#define R15B 111
#define RBP 112
#define EBP 116
#define BP 118
#define BPL 119
#define RSI 120
#define ESI 124
#define SI 126
#define SIL 127
#define RDI 128
#define EDI 132
#define DI 134
#define DIL 135
#define RSP 136
#define ESP 140
#define SP 142
#define SPL 143
#define CS 144
#define SS 146
#define DS 148
#define ES 150
#define FS 152
#define GS 154
#define RFLAGS 160

#define EFLAGS 164

#define RIP 168

#define EIP 172

#define MM0 192
#define MM1 208
#define MM2 224
#define MM3 240
#define MM4 256
#define MM5 272
#define MM6 288
#define MM7 304
#define ST0 320
#define ST1 336
#define ST2 352
#define ST3 368
#define ST4 384
#define ST5 400
#define ST6 416
#define ST7 432
#define XMM0 448
#define XMM1 464
#define XMM2 480
#define XMM3 496
#define XMM4 512
#define XMM5 528
#define XMM6 544
#define XMM7 560
#define XMM8 576
#define XMM9 592
#define XMM10 608
#define XMM11 624
#define XMM12 640
#define XMM13 656
#define XMM14 672
#define XMM15 688

#define YMM0 704
#define YMM1 736
#define YMM2 768
#define YMM3 800
#define YMM4 832
#define YMM5 864
#define YMM6 896
#define YMM7 928
#define YMM8 960
#define YMM9 992
#define YMM10 1024
#define YMM11 1056
#define YMM12 1088
#define YMM13 1120
#define YMM14 1152
#define YMM15 1184

#define FPCW 1216
#define FPSW 1218
#define FPTAG_FULL 1220
#define FPTAG 1222
#define FPIP_OFF 1224
#define FPIP_SEL 1228
#define FPOPCODE 1232
#define FPDP_OFF 1236
#define FPDP_SEL 1240

#define CR0 1244
#define CR1 1248
#define CR2 1252
#define CR3 1256
#define CR4 1260

#define MXCSR 1264
#define MXCSRMASK 1268


//#define YMM16 1216
 


  //  static unsigned char started=0;
const unsigned int register_number[NBREGS]={RAX, EAX, AX, AH, AL, RBX, EBX, BX, BH, BL,
					RCX, ECX, CX, CH, CL, RDX, EDX, DX, DH, DL,
					R8, R8D, R8W, R8B, R9, R9D, R9W, R9B,
					R10, R10D, R10W, R10B, R11, R11D, R11W, R11B,
					R12, R12D, R12W, R12B, R13, R13D, R13W, R13B,
					R14, R14D, R14W, R14B, R15, R15D, R15W, R15B,
					RBP, EBP, BP, BPL, RSI, ESI, SI, SIL,
					RDI, EDI, DI, DIL, RSP, ESP, SP, SPL,
					CS, SS, DS, ES, FS, GS, RFLAGS, RIP,
					MM0, MM1, MM2, MM3, MM4, MM5, MM6, MM7,
					ST0, ST1, ST2, ST3, ST4, ST5, ST6, ST7,
					XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6, XMM7,
					    XMM8, XMM9, XMM10, XMM11, XMM12, XMM13, XMM14, XMM15,
					    YMM0, YMM1, YMM2, YMM3, YMM4, YMM5, YMM6, YMM7,
					    YMM8, YMM9, YMM10, YMM11, YMM12, YMM13, YMM14, YMM15
};
const unsigned int register_length[NBREGS]={8, 4, 2, 1, 1, 8, 4, 2, 1, 1,
					8, 4, 2, 1, 1, 8, 4, 2, 1, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					8, 4, 2, 1, 8, 4, 2, 1,
					2, 2, 2, 2, 2, 2, 8, 8,
					16, 16, 16, 16, 16, 16, 16, 16,
					16, 16, 16, 16, 16, 16, 16, 16,
					16, 16, 16, 16, 16, 16, 16, 16,
					    16, 16, 16, 16, 16, 16, 16, 16,
					    32, 32, 32, 32, 32, 32, 32, 32,
					    32, 32, 32, 32, 32, 32, 32, 32
};


const char *regx862string(UINT32 i, UINT32 len);

UINT32 reg2int(REG r, UINT32 *len);
  
//struct Register {UINT32 num; UINT32 len;}; 
class Register 
{
 public: 
  Register();
  UINT32 num; 
  UINT32 len;
}; 

class Instruction
{
public:
  static const unsigned int MAXRRDEP=8;
  static const unsigned int MAXWRDEP=8;

  Instruction();
  Instruction(ADDRINT addr, bool is_control, UINT32 op, 
	      uint32_t nr, uint32_t nw, Register rreg[MAXRRDEP], Register wreg[MAXWRDEP],
	      ADDRINT amr, UINT32 rsize, 
	      ADDRINT amr2,
	      ADDRINT amw, UINT32 wsize,
	      xed_category_enum_t xed_category
	      );
  void dump(ostream &os);
  friend ostream &operator<<(ostream &os, const Instruction &i);
public:

  ADDRINT address;
  bool is_control;
  UINT32 opcode;
  xed_category_enum_t category;
  //private:
  uint32_t nbr;
  uint32_t nbw;
  Register rr[MAXRRDEP];
  Register wr[MAXWRDEP];
  ADDRINT amr;
  UINT32 rsize;
  ADDRINT amr2, amw;
  UINT32 wsize;
  xed_category_enum_t xed_category;

};

#endif // ILPCOMMON_HH
