#include "ilptool_interface.hh"

#include "extern.hh"

IlptoolInterface::IlptoolInterface(const char *nickname, uint32_t dh, uint32_t cid): 
  name(nickname), call_depth(dh), call_id(cid){}

IlptoolInterface::~IlptoolInterface() {}

RoutineInterface::RoutineInterface(const char *name, const char *image, ADDRINT addr, RTN rtn, UINT64 rncnt, UINT64 dephcnt, UINT64 startdepth, IlptoolInterface *ilp):
  rtn_name(name),
  rtn_image(image),
  rtn_address(addr),
  rtn_rtn(rtn),
  rtn_rtnCount(0),
  rtn_depth_counter(0),
  rtn_start_depth(0),
  is_rtn_starting_instrumentation(false)
{}

RoutineInterface::~RoutineInterface() {}


bool RoutineInterface::ctrl_start() 
{
  // We ignore ".plt" functions...
  if (this->rtn_name.compare(".plt") != 0)
    {
      call_id++;
      local_cid = call_id;

      if ( 
	  (// IF user want to dump all functions  
	   (options.option_list["dump_control"]=="all") ||
	   // OR user want to dump main image functions
	   ( (options.option_list["dump_control"]=="main") && (rtn_image == MainImageName) )
	   )
	  && // AND user depth is greater than current depth
	  (
	   ( (uint32_t)atoi(options.option_list["depth"].c_str()) > call_depth-rtn_start_depth )
	   )
	   )
	{
	  //call_id++;

	  rtn_depth_counter++;
	  call_depth++;

	  return true;
	}
    }
  return false;
}

bool RoutineInterface::ctrl_stop() 
{
  bool ret = ( 
	      (// IF user want to dump all functions
	       (options.option_list["dump_control"]=="all") ||
	       // OR user want to dump main image functions
	       ( (options.option_list["dump_control"]=="main") && (rtn_image == MainImageName) )
	       )
	       &&//|| is_instrument_on) &&
	       ( ((uint32_t)atoi(options.option_list["depth"].c_str()) > call_depth-rtn_start_depth) )
	      );
  if (ret)
    {
      call_depth--;
      rtn_depth_counter--;
      if (is_rtn_starting_instrumentation && (rtn_depth_counter <= 0) )
	{
	  is_instrument_on = false;
	}
    }
  return ret;
}

uint32_t RoutineInterface::get_depth() 
{
  return call_depth;
}

uint32_t RoutineInterface::call_depth = 0;
uint32_t RoutineInterface::call_id = 0;
bool RoutineInterface::is_instrument_on = false;
