#ifndef ILPTOOL_INTERFACE_HH
#define ILPTOOL_INTERFACE_HH

#include "ilpcommon.hh"

class IlptoolInterface
{
public:
  IlptoolInterface(const char *nickname, uint32_t depth, uint32_t call_id);
  virtual void instrument(INS ins, VOID *v)=0;
  virtual void analyse(Instruction &instr)=0;
  virtual void dump(ostream & os)=0;
  virtual void dumpxml(ostream & os, int depth)=0;
  virtual void stop()=0;
  virtual ~IlptoolInterface();
  IlptoolInterface *next_ilp;
  string name;
  uint32_t call_depth;
  uint32_t call_id;  
};

class RoutineInterface
{
public:
  RoutineInterface(const char *name, const char *image, ADDRINT addr, RTN rtn, UINT64 rncnt, UINT64 dephcnt, UINT64 startdepth, IlptoolInterface *ilp);
  virtual void start()=0;
  virtual void stop()=0;
  virtual ~RoutineInterface();
  
  // Control functions
  bool ctrl_start();
  bool ctrl_stop();
  uint32_t get_depth();

protected:
  // ROUTINE INFORMATION
  string rtn_name;
  string rtn_image;
  ADDRINT rtn_address;
  RTN rtn_rtn;
  UINT64 rtn_rtnCount;
  UINT64 rtn_depth_counter;
  UINT64 rtn_start_depth;
  bool is_rtn_starting_instrumentation;

  // STATIC (global) INFORMATION
public:
  uint32_t local_cid;
  static uint32_t call_depth;
  static uint32_t call_id;
  static bool is_instrument_on;

};

#endif // ILPTOOL_INTERFACE_HH
