#include "instrument.hh"

#include "tools/ilpprofile.hh"
#include "tools/ilphistogram.hh"
#include "tools/ilptrace.hh"
#include "extern.hh"
#include "sstream"


const char * StripPath(const char * path)
{
    const char * file = strrchr(path,'/');
    if (file)
        return file+1;
    else
        return path;
}

// INSTRUMENTATION FUNCTION
VOID InstrumentAnalyse(IlptoolInterface *ilpobj, ADDRINT a, BOOL is_control, UINT32 op, UINT32 nr, UINT32 nw,
    UINT32 rr0, UINT32 lenrr0,
    UINT32 rr1, UINT32 lenrr1,
    UINT32 rr2, UINT32 lenrr2,
    UINT32 rr3, UINT32 lenrr3,
    UINT32 rr4, UINT32 lenrr4,
    UINT32 rr5, UINT32 lenrr5,
    UINT32 rr6, UINT32 lenrr6,
    UINT32 rr7, UINT32 lenrr7,
    UINT32 wr0, UINT32 lenwr0,
    UINT32 wr1, UINT32 lenwr1,
    UINT32 wr2, UINT32 lenwr2,
    UINT32 wr3, UINT32 lenwr3,
    UINT32 wr4, UINT32 lenwr4,
    UINT32 wr5, UINT32 lenwr5,
    UINT32 wr6, UINT32 lenwr6,
    UINT32 wr7, UINT32 lenwr7,
		       ADDRINT amr, UINT32 rsize, ADDRINT amr2, ADDRINT amw, UINT32 wsize, UINT32 xed_category)
{
  if (ilpobj)
    {
  Register rr[8];
  Register wr[8];
  rr[0].num = rr0;
  rr[1].num = rr1;
  rr[2].num = rr2;
  rr[3].num = rr3;
  rr[4].num = rr4;
  rr[5].num = rr5;
  rr[6].num = rr6;
  rr[7].num = rr7;
  rr[0].len = lenrr0;
  rr[1].len = lenrr1;
  rr[2].len = lenrr2;
  rr[3].len = lenrr3;
  rr[4].len = lenrr4;
  rr[5].len = lenrr5;
  rr[6].len = lenrr6;
  rr[7].len = lenrr7;
  wr[0].num = wr0;
  wr[1].num = wr1;
  wr[2].num = wr2;
  wr[3].num = wr3;
  wr[4].num = wr4;
  wr[5].num = wr5;
  wr[6].num = wr6;
  wr[7].num = wr7;
  wr[0].len = lenwr0;
  wr[1].len = lenwr1;
  wr[2].len = lenwr2;
  wr[3].len = lenwr3;
  wr[4].len = lenwr4;
  wr[5].len = lenwr5;
  wr[6].len = lenwr6;
  wr[7].len = lenwr7;
  Instruction instr(a, (bool)is_control, op, nr, nw, rr, wr, 
		    amr,
		    rsize,
		    amr2,
		    amw,
		    wsize,
		    (xed_category_enum_t)xed_category
		    );
  ilpobj->analyse(instr);
    }
}
// ILP COMPUTATION  



VOID startacall(RoutineInterface *rtn)
{
  rtn->start();
}

VOID stopacall(RoutineInterface *rtn)
{
  rtn->stop();
}
   

// Pin calls this function every time a new rtn is executed
VOID InstrumentRoutine(RTN rtn, VOID *v)
{
  RoutineInterface *ri;
  if (options.option_list["command"]=="profile")
    {
      ri = new RoutineProfile(RTN_Name(rtn).c_str(), 
			      //StripPath(IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str()),
			      IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str(),
			      RTN_Address(rtn),
			      rtn,
			      main_ilp);
    }
  else if (options.option_list["command"]=="histogram")
    {
      ri = new RoutineHistogram(RTN_Name(rtn).c_str(), 
				//StripPath(IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str()),
				IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str(),
				RTN_Address(rtn),
				rtn,
				main_ilp);

    }

  else if (options.option_list["command"]=="trace")
    {
      ri = new RoutineTrace(RTN_Name(rtn).c_str(), 
			    //StripPath(IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str()),
			    IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str(),
			    RTN_Address(rtn),
			    rtn,
			    main_ilp);
    }
  else
    {
      cerr << "Bad option !!!" << endl;
      abort();
    }

  RTN_Open(rtn);
            
  RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)startacall,
		 IARG_PTR, ri,
		 IARG_END);
  
  RTN_InsertCall(rtn, IPOINT_AFTER, (AFUNPTR)stopacall,
		 IARG_PTR, ri,
		 IARG_END);
  

 
  for (INS ins = RTN_InsHead(rtn); INS_Valid(ins); ins = INS_Next(ins))
    {
      //      IlpList[0]->InstrumentInstructionTrace(ins, 0);
      IlpList[0]->instrument(ins, 0);
    }

  RTN_Close(rtn);
}

// This function is called when the application exits
// It prints the name and count for each procedure
VOID InstrumentFini(INT32 code, VOID *v)
{
  // Output on xml file
  outxml << endl;

  //  if(options.ht_all)
  if(options.option_list["inst_control"]=="all")
    {
      outfile << endl; 
      outfile << "Global ILP ";
      IlpList[0]->dump(outfile);
      outfile << endl; 

      // We need to close the _start call function which never return 
      outxml << "</call>" << endl;

      outxml << "<global>"<< endl;
      IlpList[0]->dumpxml(outxml, 0);
      outxml << "</global>"<< endl;
            
    }

  outxml << "</profile>"<<endl;
  outxml << endl; 

}

