/*BEGIN_LEGAL
Intel Open Source License

Copyright (c) 2002-2009 Intel Corporation. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

#include <locale>

#include <iostream>
#include <fstream>
#include <string>
#include "pin.H"
#include "tools/ilpprofile.hh"
#include "tools/ilphistogram.hh"
#include "tools/ilptrace.hh"

#include "instrument.hh"
  
#include "options.hh"

#include <unistd.h>

using std::cout;
using std::cerr;
using std::endl;
using std::ofstream;
using std::ifstream;

using std::string;


IlptoolInterface *main_ilp;//("GLOBAL");

string MainImageName = "";

vector<IlptoolInterface *> IlpList;

ofstream outfile;
ofstream outxml;


VOID RoutineLoad(IMG img, VOID *v)
{
  cerr << "Loading image: " << IMG_Name(img) << endl;
  if (IMG_IsMainExecutable(img)) MainImageName = IMG_Name(img).c_str();
      cerr << "\tMain image: " << IMG_Name(img) << endl;
      // Forward pass over all sections in an image
      for( SEC sec= IMG_SecHead(img); SEC_Valid(sec); sec = SEC_Next(sec) )
	{
	  if (SEC_IsExecutable(sec))
	    {
	      // Forward pass over all routines in a section
	      for( RTN rtn= SEC_RtnHead(sec); RTN_Valid(rtn); rtn = RTN_Next(rtn) )
		{
		  // Register Routine to be called to instrument rtn
		  InstrumentRoutine(rtn, v);
		}
	    }
	}
}

Options options;

// argc, argv are the entire command line, including pin -t <toolname> -- ...
int main(int argc, char * argv[])
{  
  std::locale::global(std::locale("C"));

  options.ReadXML();
  string profile_filename = options.option_list["resdir"] + ("/_ilptool.profile"); 
  cerr << "OUTFILE: " << profile_filename << endl;
  outfile.open(profile_filename.c_str());

  if (options.option_list["command"]=="profile")
    {
      main_ilp = new IlpProfile("GLOBAL", 0, 0);
      string profile_xmlfilename = options.option_list["resdir"] + ("/_ilptool_profile.xml"); 
      cerr << "OUTXML: " << profile_xmlfilename << endl;
      outxml.open(profile_xmlfilename.c_str());
      outxml << "<?xml version=\"1.0\"?>" << endl << endl;
      outxml << "<profile>" << endl;
    }

  if(options.option_list["command"]=="histogram")
    {
      main_ilp = new IlpHistogram("GLOBAL", 0, 0);
      string profile_xmlfilename = options.option_list["resdir"] + ("/_ilptool_profile_dummy_h.xml"); 
      cerr << "OUTXML: " << profile_xmlfilename << endl;
      outxml.open(profile_xmlfilename.c_str());
      outxml << "<?xml version=\"1.0\"?>" << endl << endl;
      outxml << "<profile>" << endl;
      
    }

  if(options.option_list["command"]=="trace")
    {
      main_ilp = new IlpTrace("GLOBAL", 0, 0);
      string profile_xmlfilename = options.option_list["resdir"] + ("/_ilptool_profile_dummy_t.xml"); 
      cerr << "OUTXML: " << profile_xmlfilename << endl;
      outxml.open(profile_xmlfilename.c_str());
      outxml << "<?xml version=\"1.0\"?>" << endl << endl;
      outxml << "<profile>" << endl;
    }

#if defined(TARGET_IA32E)
  cerr << "\tTARGET_IA32E (Registers for TARGET_IA32 with Extension)" << endl;
#else
  cerr << "\tTARGET_IA32 (Registers for TARGET_IA32 without Extension)" << endl;
#endif

  //  if (!options.ht_all)
  if (options.option_list["inst_control"]!="all")
    main_ilp->stop();
  IlpList.push_back(main_ilp);
  // Initialize symbol table code, needed for rtn instrumentation
  PIN_InitSymbols();
  // Initialize pin
  PIN_Init(argc, argv);

  // Register ImageLoad to be called when an image is loaded
  IMG_AddInstrumentFunction(RoutineLoad, 0);

  // Register Fini to be called when the application exits
  PIN_AddFiniFunction(InstrumentFini, 0);
  
  // Start the program, never returns
  PIN_StartProgram();
  
  return 0;
}
