#include "options.hh"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <map>

#include <libxml++/libxml++.h>
#include <libxml++/parsers/textreader.h>

using std::ifstream;
using std::cerr;
using std::endl;
using std::map;

using std::setw;


Options::Options()
{
  option_list["command"] = "None";
  option_list["resdir"] = "None";
  option_list["length"] = "None";
  option_list["depth"] = "0";
  option_list["inst_control"] = "None";
  option_list["dump_control"] = "None";
  option_list["floatonly"] = "None";
}

void Options::ReadXML()
{
  // Set the global C and C++ locale to the user-configured locale,
  // so we can use std::cout with UTF-8, via Glib::ustring, without exceptions.
  //  std::locale::global(std::locale(""));
  // NO SET IT TO "C"
  std::locale::global(std::locale("C"));

#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
  try
  {
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED
    cerr << "---------------------------------------------------------------------------------------------------------" << endl;
    cerr << "Reading command file : ..." << endl;
    cerr << "---------------------------------------------------------------------------------------------------------" << endl;
    //    cerr << " --- XML DEBUG 1 ---" << endl;
    xmlpp::TextReader reader("_ilptool_command.xml");
    //    cerr << " --- XML DEBUG 2 ---" << endl;
    reader.read();
    reader.read();

    string current_option="";
    
    while(reader.read())
    {
      switch(reader.get_node_type())
	{
	case (xmlpp::TextReader::Element):
	  //	  std::cerr << "ELEMENT: " << reader.get_name() << std::endl;
	  current_option = reader.get_name();
	  if (option_list.find(current_option) != option_list.end())
	    {
	      //	      std::cerr << "Set current option to : " << current_option << " ..." << std::endl;
	      //	      current_option = 
	    }
	  else
	    {
	      std::cerr << "Invalid option : " << current_option << "" << std::endl;	      
	      current_option="";
	    }
	  break;
	case (xmlpp::TextReader::Text):
	  //	  std::cerr << "TEXT  : " << reader.get_name() << std::endl;
	  if(reader.has_value())
	    {
	      //	      std::cout << "value: '" << reader.get_value() << "'" << std::endl;
	      if (option_list.find(current_option) != option_list.end())
		option_list[current_option] = reader.get_value();
	    }
	  else
	    //	    std::cout << "novalue" << std::endl;
	  
	default:
	  break;
	}
    }
#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
   }
  catch(const std::exception& e)
  {
    std::cout << "Exception caught: " << e.what() << std::endl;
  }
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED
  
  cerr << "Option list :" << endl;
  for (map<string, string>::iterator it=option_list.begin(); it!=option_list.end(); it++)
    {
      string cmd = "[" + it->first + "]";
      
      //      cerr << "\t"<< setw(8) << "[" << it->first << "]" << setw(8) << "  =  " << it->second << endl;    
      cerr << setw(25) << cmd.c_str() << setw(0) << "" << "  =  " << it->second << endl;    
    }

  cerr << "---------------------------------------------------------------------------------------------------------" << endl;
}

