#include "register.hh"
#include <iostream>

//to convert an x86 register number (our numbering) into a string
//the "len" parameter serves to output a 16-bits matching register name
//(len==2) or an 8-bits one (len==1)
//in our numbering, these registers share the same number
//for example, register AX represents both registers "ax" and "ah"
//the reason is that the register name (AX) serves as an address
//for the register. AH and AX are assumed to share their leading byte 
const char *regx862string(UINT32 i, UINT32 len){
    switch(i){
    case RAX: return "rax";
    case EAX: return "eax";
    case AX: if (len==2) return "ax";
    else return "ah";
    case AL: return "al";
    case RBX: return "rbx";
    case EBX: return "ebx";
    case BX: if (len==2) return "bx";
    else return "bh";
    case BL: return "bl";
    case RCX: return "rcx";
    case ECX: return "ecx";
    case CX: if (len==2) return "cx";
    else return "ch";
    case CL: return "cl";
    case RDX: return "rdx";
    case EDX: return "edx";
    case DX: if (len==2) return "dx";
    else return "dh";
    case DL: return "dl";
    case R8: return "r8";
    case R8D: return "r8d";
    case R8W: return "r8w";
    case R8B: return "r8b";
    case R9: return "r9";
    case R9D: return "r9d";
    case R9W: return "r9w";
    case R9B: return "r9b";
    case R10: return "r10";
    case R10D: return "r10d";
    case R10W: return "r10w";
    case R10B: return "r10b";
    case R11: return "r11";
    case R11D: return "r11d";
    case R11W: return "r11w";
    case R11B: return "r11b";
    case R12: return "r12";
    case R12D: return "r12d";
    case R12W: return "r12w";
    case R12B: return "r12b";
    case R13: return "r13";
    case R13D: return "r13d";
    case R13W: return "r13w";
    case R13B: return "r13b";
    case R14: return "r14";
    case R14D: return "r14d";
    case R14W: return "r14w";
    case R14B: return "r14b";
    case R15: return "r15";
    case R15D: return "r15d";
    case R15W: return "r15w";
    case R15B: return "r15b";
    case RBP: return "rbp";
    case EBP: return "ebp";
    case BP: return "bp";
    case BPL: return "bpl";
    case RSI: return "rsi";
    case ESI: return "esi";
    case SI: return "si";
    case SIL: return "sil";
    case RDI: return "rdi";
    case EDI: return "edi";
    case DI: return "di";
    case DIL: return "dil";
    case RSP: return "rsp";
    case ESP: return "esp";
    case SP: return "sp";
    case SPL: return "spl";
    case CS: return "cs";
    case SS: return "ss";
    case DS: return "ds";
    case ES: return "es";
    case FS: return "fs";
    case GS: return "gs";
    case RFLAGS: return "rflags";
    case RIP: return "rip";
    case MM0: return "mm0";
    case MM1: return "mm1";
    case MM2: return "mm2";
    case MM3: return "mm3";
    case MM4: return "mm4";
    case MM5: return "mm5";
    case MM6: return "mm6";
    case MM7: return "mm7";
    case ST0: return "st(0)";
    case ST1: return "st(1)";
    case ST2: return "st(2)";
    case ST3: return "st(3)";
    case ST4: return "st(4)";
    case ST5: return "st(5)";
    case ST6: return "st(6)";
    case ST7: return "st(7)";
    case XMM0: return "xmm0";
    case XMM1: return "xmm1";
    case XMM2: return "xmm2";
    case XMM3: return "xmm3";
    case XMM4: return "xmm4";
    case XMM5: return "xmm5";
    case XMM6: return "xmm6";
    case XMM7: return "xmm7";
//#if defined(TARGET_IA32E)
    case XMM8: return "xmm8";
    case XMM9: return "xmm9";
    case XMM10: return "xmm10";
    case XMM11: return "xmm11";
    case XMM12: return "xmm12";
    case XMM13: return "xmm13";
    case XMM14: return "xmm14";
    case XMM15: return "xmm15";
    case YMM0: return "ymm0";
    case YMM1: return "ymm1";
    case YMM2: return "ymm2";
    case YMM3: return "ymm3";
    case YMM4: return "ymm4";
    case YMM5: return "ymm5";
    case YMM6: return "ymm6";
    case YMM7: return "ymm7";
    case YMM8: return "ymm8";
    case YMM9: return "ymm9";
    case YMM10: return "ymm10";
    case YMM11: return "ymm11";
    case YMM12: return "ymm12";
    case YMM13: return "ymm13";
    case YMM14: return "ymm14";
    case YMM15: return "ymm15";
//#endif
    case MXCSR: return "mxcsr";
    case MXCSRMASK: return "mxcsrmask";
    case FPCW: return "fpcw";
    case FPSW: return "fpsw";
    case FPTAG_FULL: return "fptag_full";
    case FPTAG: return "fptag";
    case FPIP_OFF: return "fpip_off";
    case FPIP_SEL: return "fpip_sel";
    case FPOPCODE: return "fpopcode";
    case FPDP_OFF: return "fpdp_off";
    case FPDP_SEL: return "fpdp_sel";
    case CR0: return "cr0";
    case CR1: return "cr1";
    case CR2: return "cr2";
    case CR3: return "cr3";
    case CR4: return "cr4";
    default: return "unknown";
    }
  }

//to convert from our numbering to PIN numbering
REG regx862regpin(UINT32 i){
    switch(i){
    case RAX:
    case EAX: 
    case AX:
    case AL: 
#if defined(TARGET_IA32E)
              return REG_RAX;
#else
              return REG_EAX;
#endif
    case RBX: 
    case EBX: 
    case BX: 
    case BL: 
#if defined(TARGET_IA32E)
              return REG_RBX;
#else
              return REG_EBX;
#endif
    case RCX:
    case ECX: 
    case CX: 
    case CL: 
#if defined(TARGET_IA32E)
              return REG_RCX;
#else
              return REG_ECX;
#endif
    case RDX: 
    case EDX:
    case DX: 
    case DL: 
#if defined(TARGET_IA32E)
              return REG_RDX;
#else
              return REG_EDX;
#endif
#if defined(TARGET_IA32E)
    case R8: 
    case R8D: 
    case R8W: 
    case R8B: 
              return REG_R8;
    case R9: 
    case R9D:
    case R9W: 
    case R9B: 
              return REG_R9;
    case R10: 
    case R10D: 
    case R10W: 
    case R10B: 
              return REG_R10;
    case R11: 
    case R11D: 
    case R11W: 
    case R11B: 
              return REG_R11;
    case R12: 
    case R12D: 
    case R12W: 
    case R12B: 
              return REG_R12;
    case R13: 
    case R13D: 
    case R13W:
    case R13B: 
              return REG_R13;
    case R14: 
    case R14D: 
    case R14W: 
    case R14B: 
              return REG_R14;
    case R15: 
    case R15D: 
    case R15W: 
    case R15B: 
              return REG_R15;
#endif
    case RBP:
    case EBP: 
    case BP: 
    case BPL: 
#if defined(TARGET_IA32E)
              return REG_RBP;
#else
              return REG_EBP;
#endif
    case RSI: 
    case ESI: 
    case SI: 
    case SIL: 
#if defined(TARGET_IA32E)
              return REG_RSI;
#else
              return REG_ESI;
#endif
    case RDI: 
    case EDI: 
    case DI: 
    case DIL: 
#if defined(TARGET_IA32E)
              return REG_RDI;
#else
              return REG_EDI;
#endif
    case RSP: 
    case ESP: 
    case SP: 
    case SPL: 
#if defined(TARGET_IA32E)
              return REG_RSP;
#else
              return REG_ESP;
#endif
    case CS: return REG_SEG_CS;
    case SS: return REG_SEG_SS;
    case DS: return REG_SEG_DS;
    case ES: return REG_SEG_ES;
    case FS: return REG_SEG_FS;
    case GS: return REG_SEG_GS;
    case RFLAGS: 
#if defined(TARGET_IA32E)
              return REG_RFLAGS;
#else
              return REG_EFLAGS;
#endif
    case RIP: 
#if defined(TARGET_IA32E)
              return REG_RIP;
#else
              return REG_EIP;
#endif
    case MM0: return REG_MM0;
    case MM1: return REG_MM1;
    case MM2: return REG_MM2;
    case MM3: return REG_MM3;
    case MM4: return REG_MM4;
    case MM5: return REG_MM5;
    case MM6: return REG_MM6;
    case MM7: return REG_MM7;
    case ST0: return REG_ST0;
    case ST1: return REG_ST1;
    case ST2: return REG_ST2;
    case ST3: return REG_ST3;
    case ST4: return REG_ST4;
    case ST5: return REG_ST5;
    case ST6: return REG_ST6;
    case ST7: return REG_ST7;
    case XMM0: return REG_XMM0;
    case XMM1: return REG_XMM1;
    case XMM2: return REG_XMM2;
    case XMM3: return REG_XMM3;
    case XMM4: return REG_XMM4;
    case XMM5: return REG_XMM5;
    case XMM6: return REG_XMM6;
    case XMM7: return REG_XMM7;
#if defined(TARGET_IA32E)
    case XMM8: return REG_XMM8;
    case XMM9: return REG_XMM9;
    case XMM10: return REG_XMM10;
    case XMM11: return REG_XMM11;
    case XMM12: return REG_XMM12;
    case XMM13: return REG_XMM13;
    case XMM14: return REG_XMM14;
    case XMM15: return REG_XMM15;
    case YMM0: return REG_YMM0;
    case YMM1: return REG_YMM1;
    case YMM2: return REG_YMM2;
    case YMM3: return REG_YMM3;
    case YMM4: return REG_YMM4;
    case YMM5: return REG_YMM5;
    case YMM6: return REG_YMM6;
    case YMM7: return REG_YMM7;
    case YMM8: return REG_YMM8;
    case YMM9: return REG_YMM9;
    case YMM10: return REG_YMM10;
    case YMM11: return REG_YMM11;
    case YMM12: return REG_YMM12;
    case YMM13: return REG_YMM13;
    case YMM14: return REG_YMM14;
    case YMM15: return REG_YMM15;
#endif
    case MXCSR: return REG_MXCSR;
    case MXCSRMASK: return REG_MXCSRMASK;
    case FPCW: return REG_FPCW;
    case FPSW: return REG_FPSW;
    case FPTAG_FULL: return REG_FPTAG_FULL;
    case FPTAG: return REG_FPTAG;
    case FPIP_OFF: return REG_FPIP_OFF;
    case FPIP_SEL: return REG_FPIP_SEL;
    case FPOPCODE: return REG_FPOPCODE;
    case FPDP_OFF: return REG_FPDP_OFF;
    case FPDP_SEL: return REG_FPDP_SEL;
    case CR0: return REG_CR0;
    case CR1: return REG_CR1;
    case CR2: return REG_CR2;
    case CR3: return REG_CR3;
    case CR4: return REG_CR4;
    default: return REG_INVALID_;
    }
}

//to convert from PIN numbering to our numbering
//the main function result is the register number in our numbering system
//the side effect result "len" is the size in bytes of the requested register
UINT32 reg2int(REG r, UINT32 *len){
    unsigned int reg;
    reg = (unsigned int)r;
    switch(r){
    case REG_INVALID_: *len=10000; return 10000;
#if defined(TARGET_IA32E)
    case REG_RDI: *len=8; return RDI;//rdi
    case REG_RSI: *len=8; return RSI;//rsi
    case REG_RBP: *len=8; return RBP;//rbp
    case REG_RSP: *len=8; return RSP;//rsp
    case REG_RBX: *len=8; return RBX;//rbx
    case REG_RDX: *len=8; return RDX;//rdx
    case REG_RCX: *len=8; return RCX;//rcx
    case REG_RAX: *len=8; return RAX;//rax
    case REG_R8: *len=8; return R8;//r8
    case REG_R9: *len=8; return R9;//r9
    case REG_R10: *len=8; return R10;//r10
    case REG_R11: *len=8; return R11;//r11
    case REG_R12: *len=8; return R12;//r12
    case REG_R13: *len=8; return R13;//r13
    case REG_R14: *len=8; return R14;//r14
    case REG_R15: *len=8; return R15;//r15
    case REG_RFLAGS: *len=8; return RFLAGS;//rflags
    case REG_RIP: *len=8; return RIP;//rip
#endif
    case REG_EFLAGS: *len=8; return RFLAGS;//rflags
    case REG_EIP: *len=8; return RIP;//rip      
    case REG_SEG_CS: *len=2; return CS;//cs
    case REG_SEG_SS: *len=2; return SS;//ss
    case REG_SEG_DS: *len=2; return DS;//ds
    case REG_SEG_ES: *len=2; return ES;//es
    case REG_SEG_FS: *len=2; return FS;//fs
    case REG_SEG_GS: *len=2; return GS;//gs
    case REG_AL: *len=1; return AL;//al
    case REG_AH: *len=1; return AH;//ah
    case REG_AX: *len=2; return AX;//ax
    case REG_CL: *len=1; return CL;//cl
    case REG_CH: *len=1; return CH;//ch
    case REG_CX: *len=2; return CX;//cx
    case REG_DL: *len=1; return DL;//dl
    case REG_DH: *len=1; return DH;//dh
    case REG_DX: *len=2; return DX;//dx
    case REG_BL: *len=1; return BL;//bl
    case REG_BH: *len=1; return BH;//bh
    case REG_BX: *len=2; return BX;//bx
    case REG_BP: *len=2; return BP;//bp
    case REG_SI: *len=2; return SI;//si
    case REG_DI: *len=2; return DI;//di
    case REG_SP: *len=2; return SP;//sp
    case REG_FLAGS: *len=8; return RFLAGS;//rflags
    case REG_IP: *len=8; return RIP;//rip
    case REG_EDI: *len=4; return EDI;//edi
#if defined(TARGET_IA32E)
    case REG_DIL: *len=1; return DIL;//dil
#endif
    case REG_ESI: *len=4; return ESI;//esi
#if defined(TARGET_IA32E)
    case REG_SIL: *len=1; return SIL;//sil
#endif
    case REG_EBP: *len=4; return EBP;//ebp
#if defined(TARGET_IA32E)
    case REG_BPL: *len=1; return BPL;//bpl
#endif
    case REG_ESP: *len=4; return ESP;//esp
#if defined(TARGET_IA32E)
    case REG_SPL: *len=1; return SPL;//spl
#endif
    case REG_EBX: *len=4; return EBX;//ebx
    case REG_EDX: *len=4; return EDX;//edx
    case REG_ECX: *len=4; return ECX;//ecx
    case REG_EAX: *len=4; return EAX;//eax
#if defined(TARGET_IA32E)
    case REG_R8B: *len=1; return R8B;//r8b
    case REG_R8W: *len=2; return R8W;//r8w
    case REG_R8D: *len=4; return R8D;//r8d
    case REG_R9B: *len=1; return R9B;//r9b
    case REG_R9W: *len=2; return R9W;//r9w
    case REG_R9D: *len=4; return R9D;//r9d
    case REG_R10B: *len=1; return R10B;//r10b
    case REG_R10W: *len=2; return R10W;//r10w
    case REG_R10D: *len=4; return R10D;//r10d
    case REG_R11B: *len=1; return R11B;//r11b
    case REG_R11W: *len=2; return R11W;//r11w
    case REG_R11D: *len=4; return R11D;//r11d
    case REG_R12B: *len=1; return R12B;//r12b
    case REG_R12W: *len=2; return R12W;//r12w
    case REG_R12D: *len=4; return R12D;//r12d
    case REG_R13B: *len=1; return R13B;//r13b
    case REG_R13W: *len=2; return R13W;//r13w
    case REG_R13D: *len=4; return R13D;//r13d
    case REG_R14B: *len=1; return R14B;//r14b
    case REG_R14W: *len=2; return R14W;//r14w
    case REG_R14D: *len=4; return R14D;//r14d
    case REG_R15B: *len=1; return R15B;//r15b
    case REG_R15W: *len=2; return R15W;//r15w
    case REG_R15D: *len=4; return R15D;//r15d
#endif
    case REG_MM0: *len=16; return MM0;//mm0
    case REG_MM1: *len=16; return MM1;//mm1
    case REG_MM2: *len=16; return MM2;//mm2
    case REG_MM3: *len=16; return MM3;//mm3
    case REG_MM4: *len=16; return MM4;//mm4
    case REG_MM5: *len=16; return MM5;//mm5
    case REG_MM6: *len=16; return MM6;//mm6
    case REG_MM7: *len=16; return MM7;//mm7
    case REG_XMM0: *len=16; return XMM0;//xmm0
    case REG_XMM1: *len=16; return XMM1;//xmm1
    case REG_XMM2: *len=16; return XMM2;//xmm2
    case REG_XMM3: *len=16; return XMM3;//xmm3
    case REG_XMM4: *len=16; return XMM4;//xmm4
    case REG_XMM5: *len=16; return XMM5;//xmm5
    case REG_XMM6: *len=16; return XMM6;//xmm6
    case REG_XMM7: *len=16; return XMM7;//xmm7
 
#if defined(TARGET_IA32E)
    case REG_XMM8: *len=16; return XMM8;//xmm8
    case REG_XMM9: *len=16; return XMM9;//xmm9
    case REG_XMM10: *len=16; return XMM10;//xmm10
    case REG_XMM11: *len=16; return XMM11;//xmm11
    case REG_XMM12: *len=16; return XMM12;//xmm12
    case REG_XMM13: *len=16; return XMM13;//xmm13
    case REG_XMM14: *len=16; return XMM14;//xmm14
    case REG_XMM15: *len=16; return XMM15;//xmm15

    case REG_YMM0: *len=32; return YMM0;//ymm0
    case REG_YMM1: *len=32; return YMM1;//ymm0
    case REG_YMM2: *len=32; return YMM2;//ymm0
    case REG_YMM3: *len=32; return YMM3;//ymm0
    case REG_YMM4: *len=32; return YMM4;//ymm0
    case REG_YMM5: *len=32; return YMM5;//ymm0
    case REG_YMM6: *len=32; return YMM6;//ymm0
    case REG_YMM7: *len=32; return YMM7;//ymm0
    case REG_YMM8: *len=32; return YMM8;//ymm0
    case REG_YMM9: *len=32; return YMM9;//ymm0
    case REG_YMM10: *len=32; return YMM10;//ymm0
    case REG_YMM11: *len=32; return YMM11;//ymm0
    case REG_YMM12: *len=32; return YMM12;//ymm0
    case REG_YMM13: *len=32; return YMM13;//ymm0
    case REG_YMM14: *len=32; return YMM14;//ymm0
    case REG_YMM15: *len=32; return YMM15;//ymm0
#endif
    case REG_ST0: *len=16; return ST0;//st(0)
    case REG_ST1: *len=16; return ST1;//st(1)
    case REG_ST2: *len=16; return ST2;//st(2)
    case REG_ST3: *len=16; return ST3;//st(3)
    case REG_ST4: *len=16; return ST4;//st(4)
    case REG_ST5: *len=16; return ST5;//st(5)
    case REG_ST6: *len=16; return ST6;//st(6)
    case REG_ST7: *len=16; return ST7;//st(7)
    case REG_MXCSR: *len=4; return MXCSR;
    case REG_MXCSRMASK: *len=4; return MXCSRMASK;

    case REG_FPCW: *len=2; return FPCW;
#if defined(TARGET_IA32E)
    case 182: *len=2; return FPSW;
#else
    case 120: *len=2; return FPSW;
#endif
    case REG_FPSW: *len=2; return FPSW;
    case REG_FPTAG_FULL: *len=2; return FPTAG_FULL;
    case REG_FPTAG: *len=2; return FPTAG;
    case REG_FPIP_OFF: *len=8; return FPIP_OFF;
    case REG_FPIP_SEL: *len=4; return FPIP_SEL;
    case REG_FPOPCODE: *len=2; return FPOPCODE;
    case REG_FPDP_OFF: *len=8; return FPDP_OFF;
    case REG_FPDP_SEL: *len=4; return FPDP_SEL;
    case REG_CR0: *len=4; return CR0;
    case REG_CR1: *len=4; return CR1;
    case REG_CR2: *len=4; return CR2;
    case REG_CR3: *len=4; return CR3;
    case REG_CR4: *len=4; return CR4;
    default: *len=10000+reg; return 10000+reg;//
    }
  }

void dump_pin_reg_map(){
 cerr << "REG_INVALID_" << REG_INVALID_ << endl;
 cerr << "REG_GR_BASE" << REG_GR_BASE << endl;
 cerr << "REG_RBASE" << REG_RBASE << endl;
// cerr << "REG_GBASE" << REG_GBASE << endl;
 cerr << "REG_EDI" << REG_EDI << endl;
 cerr << "REG_GDI" << REG_GDI << endl;
 cerr << "REG_ESI" << REG_ESI << endl;
 cerr << "REG_GSI" << REG_GSI << endl;
 cerr << "REG_EBP" << REG_EBP << endl;
 cerr << "REG_GBP" << REG_GBP << endl;
 cerr << "REG_ESP" << REG_ESP << endl;
 cerr << "REG_STACK_PTR" << REG_STACK_PTR << endl;
 cerr << "REG_EBX" << REG_EBX << endl;
 cerr << "REG_GBX" << REG_GBX << endl;
 cerr << "REG_EDX" << REG_EDX << endl;
 cerr << "REG_GDX" << REG_GDX << endl;
 cerr << "REG_ECX" << REG_ECX << endl;
 cerr << "REG_GCX" << REG_GCX << endl;
 cerr << "REG_EAX" << REG_EAX << endl;
 cerr << "REG_GAX" << REG_GAX << endl;
 cerr << "REG_GR_LAST" << REG_GR_LAST << endl;
// cerr << "REG_GROT_LAST" << REG_GROT_LAST << endl;
 cerr << "REG_SEG_BASE" << REG_SEG_BASE << endl;
 cerr << "REG_SEG_CS" << REG_SEG_CS << endl;
 cerr << "REG_SEG_SS" << REG_SEG_SS << endl;
 cerr << "REG_SEG_DS" << REG_SEG_DS << endl;
 cerr << "REG_SEG_ES" << REG_SEG_ES << endl;
 cerr << "REG_SEG_FS" << REG_SEG_FS << endl;
 cerr << "REG_SEG_GS" << REG_SEG_GS << endl;
 cerr << "REG_SEG_LAST" << REG_SEG_LAST << endl;
 cerr << "REG_EFLAGS" << REG_EFLAGS << endl;
 cerr << "REG_GFLAGS" << REG_GFLAGS << endl;
 cerr << "REG_EIP" << REG_EIP << endl;
 cerr << "REG_INST_PTR" << REG_INST_PTR << endl;
 cerr << "REG_AL" << REG_AL << endl;
 cerr << "REG_AH" << REG_AH << endl;
 cerr << "REG_AX" << REG_AX << endl;
 cerr << "REG_CL" << REG_CL << endl;
 cerr << "REG_CH" << REG_CH << endl;
 cerr << "REG_CX" << REG_CX << endl;
 cerr << "REG_DL" << REG_DL << endl;
 cerr << "REG_DH" << REG_DH << endl;
 cerr << "REG_DX" << REG_DX << endl;
 cerr << "REG_BL" << REG_BL << endl;
 cerr << "REG_BH" << REG_BH << endl;
 cerr << "REG_BX" << REG_BX << endl;
 cerr << "REG_BP" << REG_BP << endl;
 cerr << "REG_SI" << REG_SI << endl;
 cerr << "REG_DI" << REG_DI << endl;
 cerr << "REG_SP" << REG_SP << endl;
 cerr << "REG_FLAGS" << REG_FLAGS << endl;
 cerr << "REG_IP" << REG_IP << endl;
 cerr << "REG_MM_BASE" << REG_MM_BASE << endl;
 cerr << "REG_MM0" << REG_MM0 << endl;
 cerr << "REG_MM1" << REG_MM1 << endl;
 cerr << "REG_MM2" << REG_MM2 << endl;
 cerr << "REG_MM3" << REG_MM3 << endl;
 cerr << "REG_MM4" << REG_MM4 << endl;
 cerr << "REG_MM5" << REG_MM5 << endl;
 cerr << "REG_MM6" << REG_MM6 << endl;
 cerr << "REG_MM7" << REG_MM7 << endl;
 cerr << "REG_MM_LAST" << REG_MM_LAST << endl;
 cerr << "REG_EMM_BASE" << REG_EMM_BASE << endl;
 cerr << "REG_EMM0" << REG_EMM0 << endl;
 cerr << "REG_EMM1" << REG_EMM1 << endl;
 cerr << "REG_EMM2" << REG_EMM2 << endl;
 cerr << "REG_EMM3" << REG_EMM3 << endl;
 cerr << "REG_EMM4" << REG_EMM4 << endl;
 cerr << "REG_EMM5" << REG_EMM5 << endl;
 cerr << "REG_EMM6" << REG_EMM6 << endl;
 cerr << "REG_EMM7" << REG_EMM7 << endl;
 cerr << "REG_EMM_LAST" << REG_EMM_LAST << endl;
 cerr << "REG_MXT" << REG_MXT << endl;
 cerr << "REG_XMM_BASE" << REG_XMM_BASE << endl;
 cerr << "REG_FIRST_FP_REG" << REG_FIRST_FP_REG << endl;
 cerr << "REG_XMM0" << REG_XMM0 << endl;
 cerr << "REG_XMM1" << REG_XMM1 << endl;
 cerr << "REG_XMM2" << REG_XMM2 << endl;
 cerr << "REG_XMM3" << REG_XMM3 << endl;
 cerr << "REG_XMM4" << REG_XMM4 << endl;
 cerr << "REG_XMM5" << REG_XMM5 << endl;
 cerr << "REG_XMM6" << REG_XMM6 << endl;
 cerr << "REG_XMM7" << REG_XMM7 << endl;
#if defined(TARGET_IA32E)
 cerr << "REG_XMM8" << REG_XMM8 << endl;
 cerr << "REG_XMM9" << REG_XMM9 << endl;
 cerr << "REG_XMM10" << REG_XMM10 << endl;
 cerr << "REG_XMM11" << REG_XMM11 << endl;
 cerr << "REG_XMM12" << REG_XMM12 << endl;
 cerr << "REG_XMM13" << REG_XMM13 << endl;
 cerr << "REG_XMM14" << REG_XMM14 << endl;
 cerr << "REG_XMM15" << REG_XMM15 << endl;
 cerr << "REG_XMM_LAST" << REG_XMM_LAST << endl;
 cerr << "REG_YMM_BASE" << REG_YMM_BASE << endl;
 cerr << "REG_YMM0" << REG_YMM0 << endl;
 cerr << "REG_YMM1" << REG_YMM1 << endl;
 cerr << "REG_YMM2" << REG_YMM2 << endl;
 cerr << "REG_YMM3" << REG_YMM3 << endl;
 cerr << "REG_YMM4" << REG_YMM4 << endl;
 cerr << "REG_YMM5" << REG_YMM5 << endl;
 cerr << "REG_YMM6" << REG_YMM6 << endl;
 cerr << "REG_YMM7" << REG_YMM7 << endl;
 cerr << "REG_YMM8" << REG_YMM8 << endl;
 cerr << "REG_YMM9" << REG_YMM9 << endl;
 cerr << "REG_YMM10" << REG_YMM10 << endl;
 cerr << "REG_YMM11" << REG_YMM11 << endl;
 cerr << "REG_YMM12" << REG_YMM12 << endl;
 cerr << "REG_YMM13" << REG_YMM13 << endl;
 cerr << "REG_YMM14" << REG_YMM14 << endl;
 cerr << "REG_YMM15" << REG_YMM15 << endl;
 cerr << "REG_YMM_LAST" << REG_YMM_LAST << endl;
#endif
 cerr << "REG_MXCSR" << REG_MXCSR << endl;
 cerr << "REG_MXCSRMASK" << REG_MXCSRMASK << endl;
// cerr << "REG_ORIG_EAX" << REG_ORIG_EAX << endl;
 cerr << "REG_ORIG_GAX" << REG_ORIG_GAX << endl;
 cerr << "REG_DR_BASE" << REG_DR_BASE << endl;
 cerr << "REG_DR0" << REG_DR0 << endl;
 cerr << "REG_DR1" << REG_DR1 << endl;
 cerr << "REG_DR2" << REG_DR2 << endl;
 cerr << "REG_DR3" << REG_DR3 << endl;
 cerr << "REG_DR4" << REG_DR4 << endl;
 cerr << "REG_DR5" << REG_DR5 << endl;
 cerr << "REG_DR6" << REG_DR6 << endl;
 cerr << "REG_DR7" << REG_DR7 << endl;
 cerr << "REG_DR_LAST" << REG_DR_LAST << endl;
 cerr << "REG_CR_BASE" << REG_CR_BASE << endl;
 cerr << "REG_CR0" << REG_CR0 << endl;
 cerr << "REG_CR1" << REG_CR1 << endl;
 cerr << "REG_CR2" << REG_CR2 << endl;
 cerr << "REG_CR3" << REG_CR3 << endl;
 cerr << "REG_CR4" << REG_CR4 << endl;
 cerr << "REG_CR_LAST" << REG_CR_LAST << endl;
 cerr << "REG_TSSR" << REG_TSSR << endl;
 cerr << "REG_LDTR" << REG_LDTR << endl;
 cerr << "REG_TR_BASE" << REG_TR_BASE << endl;
 cerr << "REG_TR3" << REG_TR3 << endl;
 cerr << "REG_TR4" << REG_TR4 << endl;
 cerr << "REG_TR5" << REG_TR5 << endl;
 cerr << "REG_TR6" << REG_TR6 << endl;
 cerr << "REG_TR7" << REG_TR7 << endl;
 cerr << "REG_TR_LAST" << REG_TR_LAST << endl;
 cerr << "REG_FPST_BASE" << REG_FPST_BASE << endl;
 cerr << "REG_FPSTATUS_BASE" << REG_FPSTATUS_BASE << endl;
 cerr << "REG_FPCW" << REG_FPCW << endl;
 cerr << "REG_FPSW" << REG_FPSW << endl;
 cerr << "REG_FPTAG" << REG_FPTAG << endl;
 cerr << "REG_FPIP_OFF" << REG_FPIP_OFF << endl;
 cerr << "REG_FPIP_SEL" << REG_FPIP_SEL << endl;
 cerr << "REG_FPOPCODE" << REG_FPOPCODE << endl;
 cerr << "REG_FPDP_OFF" << REG_FPDP_OFF << endl;
 cerr << "REG_FPDP_SEL" << REG_FPDP_SEL << endl;
 cerr << "REG_FPSTATUS_LAST" << REG_FPSTATUS_LAST << endl;
 cerr << "REG_FPTAG_FULL" << REG_FPTAG_FULL << endl;
 cerr << "REG_ST_BASE" << REG_ST_BASE << endl;
 cerr << "REG_ST0" << REG_ST0 << endl;
 cerr << "REG_ST1" << REG_ST1 << endl;
 cerr << "REG_ST2" << REG_ST2 << endl;
 cerr << "REG_ST3" << REG_ST3 << endl;
 cerr << "REG_ST4" << REG_ST4 << endl;
 cerr << "REG_ST5" << REG_ST5 << endl;
 cerr << "REG_ST6" << REG_ST6 << endl;
 cerr << "REG_ST7" << REG_ST7 << endl;
 cerr << "REG_ST_LAST" << REG_ST_LAST << endl;
 cerr << "REG_FPST_LAST" << REG_FPST_LAST << endl;
 cerr << "REG_SEG_FS_BASE" << REG_SEG_FS_BASE << endl;
 cerr << "REG_SEG_GS_BASE" << REG_SEG_GS_BASE << endl;
 cerr << "REG_INST_BASE" << REG_INST_BASE << endl;
 cerr << "REG_INST_SCRATCH_BASE" << REG_INST_SCRATCH_BASE << endl;
 cerr << "REG_INST_G0" << REG_INST_G0 << endl;
 cerr << "REG_INST_G1" << REG_INST_G1 << endl;
 cerr << "REG_INST_G2" << REG_INST_G2 << endl;
 cerr << "REG_INST_G3" << REG_INST_G3 << endl;
 cerr << "REG_INST_G4" << REG_INST_G4 << endl;
 cerr << "REG_INST_G5" << REG_INST_G5 << endl;
 cerr << "REG_INST_G6" << REG_INST_G6 << endl;
 cerr << "REG_INST_G7" << REG_INST_G7 << endl;
 cerr << "REG_INST_G8" << REG_INST_G8 << endl;
 cerr << "REG_INST_G9" << REG_INST_G9 << endl;
 cerr << "REG_INST_G10" << REG_INST_G10 << endl;
 cerr << "REG_INST_G11" << REG_INST_G11 << endl;
 cerr << "REG_INST_G12" << REG_INST_G12 << endl;
 cerr << "REG_INST_G13" << REG_INST_G13 << endl;
 cerr << "REG_INST_G14" << REG_INST_G14 << endl;
 cerr << "REG_INST_G15" << REG_INST_G15 << endl;
 cerr << "REG_INST_G16" << REG_INST_G16 << endl;
 cerr << "REG_INST_G17" << REG_INST_G17 << endl;
 cerr << "REG_INST_G18" << REG_INST_G18 << endl;
 cerr << "REG_INST_G19" << REG_INST_G19 << endl;
 cerr << "REG_INST_TOOL_LAST" << REG_INST_TOOL_LAST << endl;
 cerr << "REG_BUF_BASE0" << REG_BUF_BASE0 << endl;
 cerr << "REG_BUF_BASE1" << REG_BUF_BASE1 << endl;
 cerr << "REG_BUF_BASE2" << REG_BUF_BASE2 << endl;
 cerr << "REG_BUF_BASE3" << REG_BUF_BASE3 << endl;
 cerr << "REG_BUF_BASE4" << REG_BUF_BASE4 << endl;
 cerr << "REG_BUF_BASE5" << REG_BUF_BASE5 << endl;
 cerr << "REG_BUF_BASE6" << REG_BUF_BASE6 << endl;
 cerr << "REG_BUF_BASE7" << REG_BUF_BASE7 << endl;
 cerr << "REG_BUF_BASE8" << REG_BUF_BASE8 << endl;
 cerr << "REG_BUF_BASE9" << REG_BUF_BASE9 << endl;
 cerr << "REG_BUF_LAST" << REG_BUF_LAST << endl;
 cerr << "REG_BUF_END0" << REG_BUF_END0 << endl;
 cerr << "REG_BUF_END1" << REG_BUF_END1 << endl;
 cerr << "REG_BUF_END2" << REG_BUF_END2 << endl;
 cerr << "REG_BUF_END3" << REG_BUF_END3 << endl;
 cerr << "REG_BUF_END4" << REG_BUF_END4 << endl;
 cerr << "REG_BUF_END5" << REG_BUF_END5 << endl;
 cerr << "REG_BUF_END6" << REG_BUF_END6 << endl;
 cerr << "REG_BUF_END7" << REG_BUF_END7 << endl;
 cerr << "REG_BUF_END8" << REG_BUF_END8 << endl;
 cerr << "REG_BUF_END9" << REG_BUF_END9 << endl;
 cerr << "REG_BUF_ENDLAST" << REG_BUF_ENDLAST << endl;
 cerr << "REG_INST_SCRATCH_LAST" << REG_INST_SCRATCH_LAST << endl;
 cerr << "REG_LAST" << REG_LAST << endl;
}
