#ifndef REGISTER_HH
#define REGISTER_HH

#include "pin.H"

//a structure to hold both a register number and its length in bytes
typedef struct {UINT32 num; UINT32 len;} Register;

//to convert a register number in our numbering system into a string
//the length of the register in bytes should be provided
const char *regx862string(UINT32 i, UINT32 len);

//to convert a register number in our numbering system into a PIN register number
REG regx862regpin(UINT32 i);

//to convert a PIN register number into a number in our numbering system
//the function also provides the register length in bytes
UINT32 reg2int(REG r, UINT32 *len);

void dump_pin_reg_map();

//the total number of bytes for all the machine registers
#define REG_FILE_SIZE 2048

//our numbering of the registers
//for every register, define a constant as its number 
//shared registers should be aligned
//e.g. register RAX occupies interval ranging from 16 to 23, and has number 16
//     register EAX occupies interval ranging from 20 to 23, and has number 20
//     register  AX occupies interval ranging from 22 to 23, and has number 22
//     register  AH occupies interval ranging from 22 to 22, and has number 22
//     register  AL occupies interval ranging from 23 to 23, and has number 23
//AX and AH have the same number (22)
#define RAX 16
#define EAX 20
#define AX 22
#define AH 22
#define AL 23
#define RBX 24
#define EBX 28
#define BX 30
#define BH 30
#define BL 31
#define RCX 32
#define ECX 36
#define CX 38
#define CH 38
#define CL 39
#define RDX 40
#define EDX 44
#define DX 46
#define DH 46
#define DL 47
#define R8 48
#define R8D 52
#define R8W 54
#define R8B 55
#define R9 56
#define R9D 60
#define R9W 62
#define R9B 63
#define R10 64
#define R10D 68
#define R10W 70
#define R10B 71
#define R11 72
#define R11D 76
#define R11W 78
#define R11B 79
#define R12 80
#define R12D 84
#define R12W 86
#define R12B 87
#define R13 88
#define R13D 92
#define R13W 94
#define R13B 95
#define R14 96
#define R14D 100
#define R14W 102
#define R14B 103
#define R15 104
#define R15D 108
#define R15W 110
#define R15B 111
#define RBP 112
#define EBP 116
#define BP 118
#define BPL 119
#define RSI 120
#define ESI 124
#define SI 126
#define SIL 127
#define RDI 128
#define EDI 132
#define DI 134
#define DIL 135
#define RSP 136
#define ESP 140
#define SP 142
#define SPL 143
#define CS 144
#define SS 146
#define DS 148
#define ES 150
#define FS 152
#define GS 154
#define RFLAGS 160
#define EFLAGS 164
#define RIP 168
#define EIP 172
#define MM0 192
#define MM1 208
#define MM2 224
#define MM3 240
#define MM4 256
#define MM5 272
#define MM6 288
#define MM7 304
#define ST0 320
#define ST1 336
#define ST2 352
#define ST3 368
#define ST4 384
#define ST5 400
#define ST6 416
#define ST7 432
#define XMM0 448
#define XMM1 464
#define XMM2 480
#define XMM3 496
#define XMM4 512
#define XMM5 528
#define XMM6 544
#define XMM7 560
#define XMM8 576
#define XMM9 592
#define XMM10 608
#define XMM11 624
#define XMM12 640
#define XMM13 656
#define XMM14 672
#define XMM15 688

#define YMM0 704
#define YMM1 736
#define YMM2 768
#define YMM3 800
#define YMM4 832
#define YMM5 864
#define YMM6 896
#define YMM7 928
#define YMM8 960
#define YMM9 992
#define YMM10 1024
#define YMM11 1056
#define YMM12 1088
#define YMM13 1120
#define YMM14 1152
#define YMM15 1184

#define FPCW 1216
#define FPSW 1218
#define FPTAG_FULL 1220
#define FPTAG 1222
#define FPIP_OFF 1224
#define FPIP_SEL 1228
#define FPOPCODE 1232
#define FPDP_OFF 1236
#define FPDP_SEL 1240

#define CR0 1244
#define CR1 1248
#define CR2 1252
#define CR3 1256
#define CR4 1260

#define MXCSR 1264
#define MXCSRMASK 1268
#endif // REGISTER_HH
