#include "gootrace.hh"
#include <iomanip>
#include <sstream>


GooOperation::GooOperation(): control(false), inum(0), cycle(0), address(0L), opcode(0), nbrdep(0), rdep(), nbmdep(0), mdep(), xed_category(XED_CATEGORY_INVALID)
{
  for(unsigned int i=0; i<8; i++)
    {
      rdep[i] = 0;
    }
  for(unsigned int i=0; i<8; i++)
    {
      mdep[i] = 0;
    }
  trace_key = "0";
}

GooOperation::GooOperation(bool ctrl, uint32_t in, uint32_t cy, ADDRINT addr, UINT32 op,
		     uint32_t nrd, uint32_t *rd, uint32_t nmd, uint32_t *md,
			   xed_category_enum_t category, const char *tkey, const char *_rtn_name, uint32_t _cid, uint32_t _depth):
  control(ctrl), inum(in), cycle(cy), address(addr), opcode(op), nbrdep(nrd), rdep(), nbmdep(nmd), mdep(), 
  xed_category(category), trace_key(tkey), rtn_name(_rtn_name), cid(_cid), depth(_depth)
{
  for(unsigned int i=0; i<nrd; i++)
    {
      rdep[i] = rd[i];
    }
  for(unsigned int i=0; i<nmd; i++)
    {
      mdep[i] = md[i];
    }
}

uint32_t GooOperation::get_mdep(unsigned int i)
{
  return mdep[i];
}

uint32_t GooOperation::get_nbmdep()
{
  return nbmdep;
}

uint32_t GooOperation::get_rdep(unsigned int i)
{
  return rdep[i];
}

uint32_t GooOperation::get_nbrdep()
{
  return nbrdep;
}

uint32_t GooOperation::get_cycle()
{
  return cycle;
}

uint32_t GooOperation::get_address()
{
  return address;
}

void GooOperation::dump(ostream & os)
{
  std::stringstream extra;
  extra << rtn_name << "(" << cid << ")(" << depth << ")";
  os << "[" << rtn_name << "][cid:"<<cid<<"][d:"<<depth<<"]"
     << setw(12) << inum 
     << setw(15) << setbase(16) << address << setbase(10) 
     << setw(20) << OPCODE_StringShort(opcode)
     << setw(12) << cycle 
    //   << << rtn_name << "(" << cid << ")(" << depth << ")";
    //     << setw(30) << extra 
    // << setw(15) << setbase(16) << address << setbase(10) 
    // << setw(12) << xed_category
     << setw(12) << xed_category_enum_t2str(xed_category)
     << setw(12) << setbase(16) << opcode << setbase(10);
	  //     << setw(20) << OPCODE_StringShort(opcode);
  
  unsigned int i;
  for(i=0; i<nbrdep; i++)
    os << setw(12) << rdep[i];
  for(; i<8; i++)
    os << setw(12) << "0";
  for(i=0; i<nbmdep; i++)
    os << setw(12) << mdep[i];
  for(; i<8; i++)
    os << setw(12) << "0";
}

void GooOperation::dumpkeys(ostream & os)
{
  os << setw(12) << inum 
     << setw(12) << cycle 
     << setw(15) << setbase(16) << address << setbase(10) 
     << setw(12) << xed_category
     << setw(12) << xed_category_enum_t2str(xed_category)
     << setw(12) << setbase(16) << opcode << setbase(10) 
     << setw(20) << OPCODE_StringShort(opcode);
  
  unsigned int i;
  for(i=0; i<nbrdep; i++)
    os << setw(12) << rdep[i];
  for(; i<4; i++)
    os << setw(12) << "0";
  for(i=0; i<nbmdep; i++)
    os << setw(12) << mdep[i];
  for(; i<4; i++)
    os << setw(12) << "0";

  os << setw(50) << trace_key;

}

GooTrace::GooTrace()
{

}

void GooTrace::update_control(const uint32_t num, uint32_t *numinst_control, uint32_t *nbcycles_control)
{
  uint32_t rd, md;
  uint32_t nbrd, nbmd, c;
  unsigned int i;
  operations[num].control = true;
  c = operations[num].get_cycle();
  (*numinst_control)++;
  if (c>(*nbcycles_control)) (*nbcycles_control)=c;
  nbrd = operations[num].get_nbrdep();
  for (i=0;i<nbrd;i++){
    rd=operations[num].get_rdep(i);
    if (!(operations[rd].control)) update_control(rd,numinst_control,nbcycles_control);
  }
  nbmd = operations[num].get_nbmdep();
  for (i=0;i<nbmd;i++){
    md=operations[num].get_mdep(i);
    if (!(operations[md].control)) update_control(md,numinst_control,nbcycles_control);
  }
}

void GooTrace::update(const GooOperation &op, const uint32_t num)
{
  operations[num] = op;
}

void GooTrace::dump(ostream &os, uint32_t nbinst)
{  
  for(unsigned int i=0; i< nbinst; i++)
    {
      operations[i].dump(os);
      os << endl;
    }
}

void GooTrace::resize(uint32_t size)
{
  GooOperation op;
  operations.resize(size, op);
}

GooOperation &GooTrace::operator[](uint32_t i)
{
  return operations[i];
}
