
extern "C" {
#include "xed-interface.h"
#include "xed-immdis.h"
#include "xed-portability.h"
}

#include "pin.H"

#include "tools/ilphistogram.hh"
#include "instrument.hh"

#include <iostream>
#include <sstream>
#include <iomanip>

#include "options.hh"
#include "extern.hh"
#include <clocale>

using std::cerr;
using std::endl;
using std::stringstream;


////////////////////////////////////////////////////////////////
//  CLASS ILPHISTOGRAM 
////////////////////////////////////////////////////////////////
IlpHistogram::IlpHistogram(const char *nickname, uint32_t dh, uint32_t cid):IlptoolInterface(nickname, dh, cid), histo()
{
      unsigned long i;
      for (i=0;i<REG_FILE_SIZE;i++){
	rat[i]=0;
	nbread[i]=0;
	nbwrite[i]=0;
      }
      nbcycles=0; npg=0; numinst=0;

      memory.Reset();
      stoped = false;
      is_main = false;
      if(options.option_list["command"]=="histogram")
	{
	  histo.resize((uint32_t)atoi(options.option_list["length"].c_str())+1);
	}
}


IlpHistogram::~IlpHistogram() {}

///////////////////////////////////////////////////////////
////////// START OF /// ILP COMPUTE FUNCTIONS /////////////
///////////////////////////////////////////////////////////
//VOID IlpHistogram::global_compute(Instruction &instr)//,
VOID IlpHistogram::analyse(Instruction &instr)
{  
  uint32_t c,c1,c2,c3,cexe;
    if(!stoped)
      {
	numinst++;
	// c1=readfromreg(instr.nbr,
	// 	       instr.rr[0].num, instr.rr[0].len,
	// 	       instr.rr[1].num, instr.rr[1].len,
	// 	       instr.rr[2].num, instr.rr[2].len,
	// 	       instr.rr[3].num, instr.rr[3].len
	// 	       );
	// Read cycle from register
	// Read cycle from registers
	c1=0;
	for (unsigned int i=0;i<instr.nbr;i++){
	  for (unsigned int j=0;j<instr.rr[i].len;j++){
	    c1=((c=rat[instr.rr[i].num+j])>c1)?c:c1;
	  }
	}


//   uint32_t IlpHistogram::readfromreg(UINT32 nr,
// 			    UINT32 rr0, UINT32 lenrr0,
// 			    UINT32 rr1, UINT32 lenrr1,
// 			    UINT32 rr2, UINT32 lenrr2,
// 			    UINT32 rr3, UINT32 lenrr3){
//     uint32_t cmax, c;
//     unsigned int i, j, rr[4], len[4];
//     rr[0]=rr0; rr[1]=rr1; rr[2]=rr2; rr[3]=rr3;
//     len[0]=lenrr0; len[1]=lenrr1; len[2]=lenrr2; len[3]=lenrr3;
//     cmax=0;
//     for (i=0;i<nr;i++){
// #ifdef DEBUG1
//       fprintf(fd, "instruction %ld: read from reg %s\n", numinst, regx862string(rr[i], len[i]));
// #endif
//       for (j=0;j<len[i];j++){
//         cmax=((c=rat[rr[i]+j])>cmax)?c:cmax;
//       }
//     }
//     for (i=0;i<nr;i++){
//       for (j=0;j<len[i];j++){
// 	nbread[rr[i]+j]++;
//       }
//     }
//     return cmax;
//   }


	//	cexe=c1;
	c2=(instr.amr)?readfrommem(instr.amr, instr.rsize):0;
	c3=(instr.amr2)?readfrommem(instr.amr2, instr.rsize):0;
	cexe=(c1>c2)?c1:c2;
	cexe=(c3>cexe)?c3:cexe;
	cexe++;
	// Write to registers
	// writetoreg(cexe, instr.nbw,
	// 	   instr.opcode,
	// 	   instr.wr[0].num, instr.wr[0].len,
	// 	   instr.wr[1].num, instr.wr[1].len,
	// 	   instr.wr[2].num, instr.wr[2].len,
	// 	   instr.wr[3].num, instr.wr[3].len
	// 	   );
	// Write to registers
#ifdef DEBUG_ANALYSE
	if (instr.nbw>Instruction::MAXWRDEP)
	  {
	    cerr << "ERROR instruction have too many destinations !!!" << endl;
	    //	    cerr << instr << endl;
	    instr.dump(cerr);
	    cerr << endl;
	    abort();
	  }
#endif 
	const char *opcode;
	char oponstack, opcontrol;
	opcode = OPCODE_StringShort(instr.opcode).c_str();
	oponstack = !(strcmp(opcode,"PUSH") &&
		      strcmp(opcode,"POP") &&
		      strcmp(opcode,"CALL_NEAR") &&
		      strcmp(opcode,"RET_NEAR"));
	opcontrol = (opcode[0]=='J') || !strcmp(opcode,"CALL_NEAR")
	  || !strcmp(opcode,"RET_NEAR")
	  || !strcmp(opcode,"SYSCALL");
	for (unsigned int i=0;i<instr.nbw;i++){
#if defined(TARGET_IA32E)
	  //	  if (oponstack && (instr.wr[i].num==REG_RSP)) continue;
	  //	  if (opcontrol && (instr.wr[i].num==REG_RIP)) continue;
	  if (oponstack && (instr.wr[i].num==RSP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#else
	  if (oponstack && (instr.wr[i].num==ESP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#endif
	  for (unsigned int j=0;j<instr.wr[i].len;j++){
	    rat[instr.wr[i].num+j]=cexe;
	    //	    nbwrite[instr.wr[i].num+j]++;
	  }
	}



	if(instr.amw)
	  writetomem(cexe, instr.amw, instr.wsize);
	nbcycles=(cexe>nbcycles)?cexe:nbcycles;
	//	if (options.histogram)
	if (options.option_list["command"]=="histogram")
	  histo.update(cexe,instr.xed_category);
      }
    if (next_ilp) next_ilp->analyse(instr);

}
///////////////////////////////////////////////////////////
////////// END OF /// ILP COMPUTE FUNCTIONS ///////////////
///////////////////////////////////////////////////////////

  // Pin calls this function every time a new instruction is encountered
VOID IlpHistogram::instrument(INS ins, VOID *v)
    {
      UINT32 nr, nw, i;
      OPCODE op;
      Register rr[8], wr[8];
      IlpHistogram *this_or_next = this;

      bool is_float=false;
/*
      if ( INS_RepPrefix(ins) || INS_RepnePrefix(ins) )
	{
	  // WARNING DO NOT USE REP Prefixed INSTRUCTION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
	  xed_decoded_inst_t* xedd = INS_XedDec(ins);
	  //	  xed_operand_values_t* xedv = INS_xed_operand_values(ins);
	  xed_operand_values_t* xedv = xed_decoded_inst_operands(xedd);
	  bool real_rep = xed_operand_values_has_real_rep(xedv);
	  if (real_rep)
	    {
	      // you have a 'real' rep prefix
	      cerr << "REP prefix detected and not supported yet !!!" << endl;
	      cerr << "Instruction address: " << hex << a << dec << endl;
	      cerr << INS_Disassemble(ins).c_str() << endl;
	      abort();
	    }
	}
*/

      op = INS_Opcode(ins);
      nr = INS_MaxNumRRegs(ins);

      if(nr>8) { cerr << "NR > 8 !!!" <<endl; abort(); }
      for (i=0; i<nr; i++){
	    rr[i].num = reg2int(INS_RegR(ins, i), &rr[i].len);
	    if (REG_is_fr(INS_RegR(ins, i))) is_float=true;
	    if (rr[i].num > 10000) { cerr << "RR > 10000 !!!" <<endl; abort(); }
      }
      nw = INS_MaxNumWRegs(ins);

      if(nw>8) { cerr << "NW > 8 !!!" <<endl; abort(); }
      for (i=0; i<nw; i++){
	wr[i].num = reg2int(INS_RegW(ins, i), &wr[i].len);
	if (REG_is_fr(INS_RegW(ins, i))) is_float=true;
	if (wr[i].num > 10000) 
	  { 
	    cerr << "WR > 10000 !!!: " << wr[i].num << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Mnemonic(ins) << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Disassemble(ins) << endl;
	    abort();
	  }
	
      }

     bool is_control;
     if (INS_IsBranchOrCall(ins)) is_control=true; else is_control=false;

      UINT32 xed_category = INS_Category(ins);

      if ((options.option_list["floatonly"]=="0") || is_float)
	{
	  if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins) && INS_IsMemoryWrite(ins))
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	  else if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins))
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
                           IARG_ADDRINT, 0L,
                           IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else if (INS_IsMemoryRead(ins) && INS_IsMemoryWrite(ins))
	  INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else if (INS_IsMemoryRead(ins))
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
		       IARG_ADDRINT, 0L,
		       IARG_ADDRINT, 0L,
		       IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);


	else if (INS_IsMemoryWrite(ins))
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
		       IARG_UINT32, wr[3].len,
		       IARG_ADDRINT, 0L,
		       IARG_UINT32, 0,
		       IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else 
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_ADDRINT, 0L,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);

	}
}

void IlpHistogram::dump(ostream &os)
{
  os << "::I[" << numinst << "]";
  os << "::C[" << nbcycles << "]";
  os << "::ILP[" << (double)numinst/(double)nbcycles << "]" ;
  //  if(options.histogram)
  if(options.option_list["command"]=="histogram")
    {
      //	  static int version = 0;
	  stringstream filename;
	  filename.imbue(std::locale("C"));
	  filename << options.option_list["resdir"] << "/";
	  filename << IlptoolInterface::name << "_histo_depth_" << RoutineHistogram::call_depth+1 << "_cid_" << call_id << "_ilptool.histogram" ;
	  stringstream filename2;
	  filename2.imbue(std::locale("C"));
	  filename2 << options.option_list["resdir"] << "/";
	  filename2 << IlptoolInterface::name << "_histo_depth_" << RoutineHistogram::call_depth+1 << "_cid_" << call_id << "_ilptool.histogram.colmap" ;
	  ofstream ofile(filename.str().c_str());
	  ofstream ofile2(filename2.str().c_str());
	  ofile.imbue(std::locale("C"));
	  ofile2.imbue(std::locale("C"));
	  histo.dump(ofile, nbcycles, ofile2);
	  ofile.close();
	  ofile2.close();
#ifdef SHORT_HISTO
	  stringstream sfilename;
	  sfilename << options.option_list["resdir"] << "/";
	  sfilename << name << "_shorthisto_depth_" << RoutineHistogram::call_depth+1 << "_cid_" << call_id << "_ilptool.shorthistogram" ;
	  ofstream sofile(sfilename.str().c_str());
	  histo.sdump(sofile, nbcycles);
	  sofile.close();
#endif
    }
}

void IlpHistogram::dumpxml(ostream &os, int depth)
{
}

void IlpHistogram::stop()
{
  stoped = true;
}

////////////////////////////////////////////////////////////////
//            END INTERFACE
////////////////////////////////////////////////////////////////

  
// VOID IlpHistogram::writetoreg(uint32_t c, UINT32 nw,
// 		  UINT32 op,
// 		  UINT32 wr0, UINT32 lenwr0,
// 		  UINT32 wr1, UINT32 lenwr1,
// 		  UINT32 wr2, UINT32 lenwr2,
// 		  UINT32 wr3, UINT32 lenwr3){
//     unsigned int i, j, wr[4], len[4];
//     const char *opcode;
//     char oponstack, opcontrol;
//     wr[0]=wr0; wr[1]=wr1; wr[2]=wr2; wr[3]=wr3;
//     len[0]=lenwr0; len[1]=lenwr1; len[2]=lenwr2; len[3]=lenwr3;
//     opcode = OPCODE_StringShort(op).c_str();
//     oponstack = !(strcmp(opcode,"PUSH") &&
//                   strcmp(opcode,"POP") &&
//                   strcmp(opcode,"CALL_NEAR") &&
//                   strcmp(opcode,"RET_NEAR"));
//     opcontrol = (opcode[0]=='J') || !strcmp(opcode,"CALL_NEAR")
//       || !strcmp(opcode,"RET_NEAR")
//       || !strcmp(opcode,"SYSCALL");
//     for (i=0;i<nw;i++){
//       //      if (oponstack && (wr[i]==RSP)) continue;
//       //      if (opcontrol && (wr[i]==RIP)) continue;
// #if defined(TARGET_IA32E)
// 	  //	  if (oponstack && (instr.wr[i].num==REG_RSP)) continue;
// 	  //	  if (opcontrol && (instr.wr[i].num==REG_RIP)) continue;
// 	  if (oponstack && (wr[i]==RSP)) continue;
// 	  if (opcontrol && (wr[i]==RIP)) continue;
// #else
// 	  if (oponstack && (wr[i]==ESP)) continue;
// 	  if (opcontrol && (wr[i]==RIP)) continue;
// #endif

//       for (j=0;j<len[i];j++){
//         rat[wr[i]+j]=c;
//         nbwrite[wr[i]+j]++;
//       }
//     }
//   }

//   uint32_t IlpHistogram::readfromreg(UINT32 nr,
// 			    UINT32 rr0, UINT32 lenrr0,
// 			    UINT32 rr1, UINT32 lenrr1,
// 			    UINT32 rr2, UINT32 lenrr2,
// 			    UINT32 rr3, UINT32 lenrr3){
//     uint32_t cmax, c;
//     unsigned int i, j, rr[4], len[4];
//     rr[0]=rr0; rr[1]=rr1; rr[2]=rr2; rr[3]=rr3;
//     len[0]=lenrr0; len[1]=lenrr1; len[2]=lenrr2; len[3]=lenrr3;
//     cmax=0;
//     for (i=0;i<nr;i++){
// #ifdef DEBUG1
//       fprintf(fd, "instruction %ld: read from reg %s\n", numinst, regx862string(rr[i], len[i]));
// #endif
//       for (j=0;j<len[i];j++){
//         cmax=((c=rat[rr[i]+j])>cmax)?c:cmax;
//       }
//     }
//     for (i=0;i<nr;i++){
//       for (j=0;j<len[i];j++){
// 	nbread[rr[i]+j]++;
//       }
//     }
//     return cmax;
//   }
  
  uint32_t IlpHistogram::readfrommem(ADDRINT amr, UINT32 rsize){
    return memory.ReadMemory(amr,rsize);
  }
  
  VOID IlpHistogram::writetomem(uint32_t cexe, ADDRINT amw, UINT32 wsize){
    memory.WriteMemory(cexe, amw, wsize);
  }


////////////////////////////////////////////////////////////////
//  CLASS ROUTINEPROFILE 
////////////////////////////////////////////////////////////////
RoutineHistogram::RoutineHistogram(const char *nname, const char *image, ADDRINT addr, RTN rtn, IlptoolInterface *ilp):
  RoutineInterface(nname, image, addr, rtn, 0, 0, 0, NULL)
{
  rtn_name = nname;
  //  if (ilp)
  //    call_vector.push_back(ilp);
}

RoutineHistogram::~RoutineHistogram() {}

void RoutineHistogram::start()
{
    if (ctrl_start())
    {
      IlptoolInterface *new_ilp = new IlpHistogram(this->rtn_name.c_str(), this->call_depth, this->call_id);
      (IlpList.back())->next_ilp = new_ilp;
      IlpList.push_back(new_ilp);

      //      call_vector.push_back(new_ilp);
      outfile << setw(call_depth*5) << "";
      outfile << "start : " << setw(1) << "<" /*<< rtn_image << ":"*/ << rtn_name << ">" << setw(10) << "   (depth: "<< call_depth << ", lcid: "<< local_cid << ")" << endl;
      // XML Output
      outxml << setw(call_depth*5) << "";
      outxml << "<call name=\""<< rtn_name
	//<< "\" call_id=\""<< call_id 
	     <<"\" >" << endl;  
      outxml << setw(call_depth*5) << "";
      //      outxml << "<call_id>" << call_id++ << "</call_id>" << endl;


      //#define DEBUG_ILP_VEC 1
#ifdef DEBUG_ILP_VEC
      cerr << "----------DEBUG_ILP_VEC--CTRL_START--Start- " << rtn_image << "|" << rtn_name <<" ---- depth: "<< call_depth << "--- rtn_s_d: "<< rtn_start_depth << endl;
  for (unsigned int i=0; i<IlpList.size(); i++)
    cerr << "    IlpList[" << i << "] = "<< IlpList[i] << "   (" << IlpList[i]->name << ")" << endl;
  cerr << endl;

  cerr << "----------DEBUG_ILP_VEC--CTRL_START--Stop------------" << endl;
#endif

      // outfile << setw(call_depth*5) << "";
      // outfile << "start : " << setw(15) << rtn_name << setw(10) << "   (depth: "<< call_depth << setw(10) << "rtn_s_d: "<< rtn_start_depth << ")" << endl;   

    }
}

void RoutineHistogram::stop()
{
  IlpHistogram *tmpilp=0;

  if(ctrl_stop())
    {
      // FIND FUNCTION which is STOP into IlpList vector.
      unsigned int i;
      unsigned int count=0;
      for (i=IlpList.size()-1; i>0; i--)
	{
	  if (rtn_name == IlpList[i]->name)
	    break;
	  count++;
	}
      
      if (count == IlpList.size()-1 )
	{
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "imp. stop ?: " << setw(1) << "<" << rtn_name << "> (cid: " 
	    //	  outfile << "imp. stop ?: " << setw(1) << "<" << (IlpList.back())->name << "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	  tmpilp = (IlpHistogram *)(IlpList.back());
	  tmpilp->dump(outfile);
	  tmpilp->dumpxml(outxml, call_depth+1);
	  outfile << endl;
	  call_depth++;
	}
      else if (count>0)
	{
	  //      unsigned int j;
	  //      if (count>0)
	  {
	  //	  for (j=IlpList.size()-1; j>i; j--)
	  for (unsigned int j=0; j<count; j++ )
	    {
	      outfile << setw((call_depth+1)*5) << "";
	      outfile << "imp. stop: " << setw(1) << "<" << (IlpList.back())->name<< "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	      
	      tmpilp = (IlpHistogram *)(IlpList.back());
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	      
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      (IlpList.back())->next_ilp = NULL;
	    }
	  //call_depth++;
	  }
	  // Perform the explicite STOP

	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";// << " rtn_s_d: "<< rtn_start_depth << ")";// << endl; 
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpHistogram *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }

	}
      else
	{
	  //      if (count)
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";// << " rtn_s_d: "<< rtn_start_depth << ")";// << endl; 
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpHistogram *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }
	}
#ifdef DEBUG_ILP_VEC
      //  cerr << "----------DEBUG_ILP_VEC---CTRL_STOP-- Start- " << rtn_name <<" -----------" << endl;
  cerr << "----------DEBUG_ILP_VEC--CTRL_STOP--Start- " << rtn_image << "|" << rtn_name <<" ---- depth: "<< call_depth << "--- rtn_s_d: "<< rtn_start_depth << endl;
//   for (unsigned int i=0; i<call_vector.size(); i++)
//     cerr << "call_vector["<< i << "] = " << call_vector[i] << endl;;
//   cerr << endl;

  for (unsigned int i=0; i<IlpList.size(); i++)
    cerr << "    IlpList[" << i << "] = "<< IlpList[i] << "   (" << IlpList[i]->name << ")"<< endl;

  //  cerr << "This routine : " << rtn_image << "|" << rtn_name << " " << endl;
  cerr << endl;
  
  

  cerr << "----------DEBUG_ILP_VEC---CTRL_STOP-- Stop ------------" << endl;
#endif

    }
  
}

