
extern "C" {
#include "xed-interface.h"
#include "xed-immdis.h"
#include "xed-portability.h"
}

#include "pin.H"

#include "tools/ilpprofile.hh"
#include "instrument.hh"

#include <iostream>
#include <sstream>
#include <iomanip>

#include "options.hh"
#include "extern.hh"

using std::cerr;
using std::endl;
using std::stringstream;

using std::setw;

//#define DEBUG_ANALYSE 1

////////////////////////////////////////////////////////////////
//  CLASS ILPPROFILE 
////////////////////////////////////////////////////////////////
IlpProfile::IlpProfile(const char *nickname, uint32_t dh, uint32_t cid): IlptoolInterface(nickname, dh, cid)//: histo()
{
      unsigned long i;
      for (i=0;i<REG_FILE_SIZE;i++){
	rat[i]=0;
	nbread[i]=0;
	nbwrite[i]=0;
      }
      nbcycles=0; npg=0; numinst=0;

      memory.Reset();
//      memory_trace.Reset();
      stoped = false;
      is_main = false;
      //name = nickname;
}


IlpProfile::~IlpProfile() {}

///////////////////////////////////////////////////////////
////////// START OF /// ILP COMPUTE FUNCTIONS /////////////
///////////////////////////////////////////////////////////
//VOID IlpProfile::global_compute(Instruction &instr)//,
VOID IlpProfile::analyse(Instruction &instr)
{  
    uint32_t c,c1,c2,c3,cexe;
    c = 0;
    if(!stoped)
      {
	numinst++;
#ifdef DEBUG_ANALYSE
	if (instr.nbr>Instruction::MAXRRDEP)
	  {
	    cerr << "ERROR instruction have too many sources !!!" << endl;
	    //	    cerr << instr << endl;
	    instr.dump(cerr);
	    cerr << endl;
	    abort();
	  }
#endif 
	// Read cycle from registers
	c1=0;
	for (unsigned int i=0;i<instr.nbr;i++){
	  for (unsigned int j=0;j<instr.rr[i].len;j++){
	    c1=((c=rat[instr.rr[i].num+j])>c1)?c:c1;
	  }
	}
       	//	cexe=c1;
	c2=(instr.amr)?readfrommem(instr.amr, instr.rsize):0;
	c3=(instr.amr2)?readfrommem(instr.amr2, instr.rsize):0;
	cexe=(c1>c2)?c1:c2;
	cexe=(c3>cexe)?c3:cexe;
	cexe++;

	// Write to registers
#ifdef DEBUG_ANALYSE
	if (instr.nbw>Instruction::MAXWRDEP)
	  {
	    cerr << "ERROR instruction have too many destinations !!!" << endl;
	    //	    cerr << instr << endl;
	    instr.dump(cerr);
	    cerr << endl;
	    abort();
	  }
#endif 
	const char *opcode;
	char oponstack, opcontrol;
	opcode = OPCODE_StringShort(instr.opcode).c_str();
	oponstack = !(strcmp(opcode,"PUSH") &&
		      strcmp(opcode,"POP") &&
		      strcmp(opcode,"CALL_NEAR") &&
		      strcmp(opcode,"RET_NEAR"));
	opcontrol = (opcode[0]=='J') || !strcmp(opcode,"CALL_NEAR")
	  || !strcmp(opcode,"RET_NEAR")
	  || !strcmp(opcode,"SYSCALL");
	for (unsigned int i=0;i<instr.nbw;i++){
#if defined(TARGET_IA32E)
	  //	  if (oponstack && (instr.wr[i].num==REG_RSP)) continue;
	  //	  if (opcontrol && (instr.wr[i].num==REG_RIP)) continue;
	  if (oponstack && (instr.wr[i].num==RSP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#else
	  if (oponstack && (instr.wr[i].num==ESP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#endif
	  for (unsigned int j=0;j<instr.wr[i].len;j++){
	    rat[instr.wr[i].num+j]=cexe;
	    //	    nbwrite[instr.wr[i].num+j]++;
	  }
	}

	if(instr.amw)
	  writetomem(cexe, instr.amw, instr.wsize);
	nbcycles=(cexe>nbcycles)?cexe:nbcycles;
      }
    if (next_ilp) next_ilp->analyse(instr);

}
///////////////////////////////////////////////////////////
////////// END OF /// ILP COMPUTE FUNCTIONS ///////////////
///////////////////////////////////////////////////////////

  // Pin calls this function every time a new instruction is encountered
VOID IlpProfile::instrument(INS ins, VOID *v)
    {
      UINT32 nr, nw, i;
      OPCODE op;
      //      struct {UINT32 num; UINT32 len;} rr[8], wr[8];
      Register rr[8], wr[8];
      //      ADDRINT a;
      //      a = INS_Address(ins);
      //      Instruction *localinstr;
      IlpProfile *this_or_next = this;

      bool is_float=false;

/*
      if ( INS_RepPrefix(ins) || INS_RepnePrefix(ins) )
	{
	  // WARNING DO NOT USE REP Prefixed INSTRUCTION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
	  xed_decoded_inst_t* xedd = INS_XedDec(ins);
	  //	  xed_operand_values_t* xedv = INS_xed_operand_values(ins);
	  xed_operand_values_t* xedv = xed_decoded_inst_operands(xedd);
	  bool real_rep = xed_operand_values_has_real_rep(xedv);
	  if (real_rep)
	    {
	      // you have a 'real' rep prefix
	      cerr << "REP prefix detected and not supported yet !!!" << endl;
	      cerr << "Instruction address: " << hex << a << dec << endl;
	      cerr << INS_Disassemble(ins).c_str() << endl;
	      abort();
	    }
	}
*/

      op = INS_Opcode(ins);
      nr = INS_MaxNumRRegs(ins);

      if(nr>8) { cerr << "NR > 8 !!!" <<endl; abort(); }
      for (i=0; i<nr; i++){
	    rr[i].num = reg2int(INS_RegR(ins, i), &rr[i].len);
	    if (REG_is_fr(INS_RegR(ins, i))) is_float=true;
	    if (rr[i].num > 10000) 
	      { 
		cerr << "RR > 10000 !!!" <<endl;
		cerr << hex << INS_Address(ins) << dec << "\t" << INS_Mnemonic(ins) << endl;
		cerr << hex << INS_Address(ins) << dec << "\t" << INS_Disassemble(ins) << endl;
		abort();
	      }
      }
      nw = INS_MaxNumWRegs(ins);

      if(nw>8) { cerr << "NW > 8 !!!" <<endl; abort(); }
      for (i=0; i<nw; i++){
	wr[i].num = reg2int(INS_RegW(ins, i), &wr[i].len);
	if (REG_is_fr(INS_RegW(ins, i))) is_float=true;
	if (wr[i].num > 10000) 
	  { 
	    cerr << "WR > 10000 !!!: " << wr[i].num << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Mnemonic(ins) << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Disassemble(ins) << endl;
	    abort();
	  }
	
      }

     bool is_control;
     if (INS_IsBranchOrCall(ins)) is_control=true; else is_control=false;

      UINT32 xed_category = INS_Category(ins);

      if ((options.option_list["floatonly"]=="0") || is_float)
	{

	  if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins) && INS_IsMemoryWrite(ins)){
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);
          }

	  else if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins)){
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);
        }

	else if (INS_IsMemoryRead(ins) && INS_IsMemoryWrite(ins)){
	  INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);
        }

	else if (INS_IsMemoryRead(ins)){
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
		           IARG_ADDRINT, 0L,
		           IARG_ADDRINT, 0L,
		           IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);
        }

	else if (INS_IsMemoryWrite(ins)){
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
        		   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
		           IARG_ADDRINT, 0L,
		           IARG_UINT32, 0,
		           IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);
        }

	else{ 
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_ADDRINT, 0L,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);
	}
	}
}

void IlpProfile::dump(ostream &os)
{
  os << "(cid: " << call_id << ") "; 
  os << "   I[" << numinst << "]";
  os << "::C[" << nbcycles << "]";
  os << "::ILP[" << (double)numinst/(double)nbcycles << "]" ;

}

void IlpProfile::dumpxml(ostream &os, int depth)
{
  os << setw((depth+1)*5) << "";
  os << "<instructions>"<< numinst << "</instructions>"<<endl;
  os << setw((depth+1)*5) << "";
  os << "<cycles>" << nbcycles << "</cycles>"<<endl;
  os << setw((depth+1)*5) << "";
  os << "<ilp>" << (double)numinst/(double)nbcycles << "</ilp>"<<endl;
}

void IlpProfile::stop()
{
  stoped = true;
}

////////////////////////////////////////////////////////////////
//            END INTERFACE
////////////////////////////////////////////////////////////////

uint32_t IlpProfile::readfrommem(ADDRINT amr, UINT32 rsize){
  return memory.ReadMemory(amr,rsize);
}

VOID IlpProfile::writetomem(uint32_t cexe, ADDRINT amw, UINT32 wsize){
  memory.WriteMemory(cexe, amw, wsize);
}


////////////////////////////////////////////////////////////////
//  CLASS ROUTINEPROFILE 
////////////////////////////////////////////////////////////////
RoutineProfile::RoutineProfile(const char *name, const char *image, ADDRINT addr, RTN rtn, IlptoolInterface *ilp):
  RoutineInterface(name, image, addr, rtn, 0, 0, 0, NULL)
{
  //  if (ilp)
    //   call_vector.push_back(ilp);
}

RoutineProfile::~RoutineProfile() {}

void RoutineProfile::start()
{
  if (ctrl_start())
    {
      IlptoolInterface *new_ilp = new IlpProfile(this->rtn_name.c_str(), this->call_depth, this->call_id);
      (IlpList.back())->next_ilp = new_ilp;
      IlpList.push_back(new_ilp);

      //      call_vector.push_back(new_ilp);
      outfile << setw(call_depth*5) << "";
      outfile << "start : " << setw(1) << "<" /*<< rtn_image << ":"*/ << rtn_name << ">" << setw(10) << "   (depth: "<< call_depth << ", lcid: "<< local_cid << ")" << endl;
      // XML Output
      outxml << setw(call_depth*5) << "";
      outxml << "<call name=\""<< rtn_name
	//<< "\" call_id=\""<< call_id 
	     <<"\" >" << endl;  
      outxml << setw(call_depth*5) << "";
      //      outxml << "<call_id>" << call_id++ << "</call_id>" << endl;


      //#define DEBUG_ILP_VEC 1
#ifdef DEBUG_ILP_VEC
      cerr << "----------DEBUG_ILP_VEC--CTRL_START--Start- " << rtn_image << "|" << rtn_name <<" ---- depth: "<< call_depth << "--- rtn_s_d: "<< rtn_start_depth << endl;
  for (unsigned int i=0; i<IlpList.size(); i++)
    cerr << "    IlpList[" << i << "] = "<< IlpList[i] << "   (" << IlpList[i]->name << ")" << endl;
  cerr << endl;

  cerr << "----------DEBUG_ILP_VEC--CTRL_START--Stop------------" << endl;
#endif

    }
}

void RoutineProfile::stop()
{
  IlpProfile *tmpilp=0;

  if (ctrl_stop())
    {
      // FIND FUNCTION which is STOP into IlpList vector.
      unsigned int i;
      unsigned int count=0;
      for (i=IlpList.size()-1; i>0; i--)
	{
	  if (rtn_name == IlpList[i]->name)
	    break;
	  count++;
	}
      
      if (count == IlpList.size()-1 )
	{
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "imp. stop ?: " << setw(1) << "<" << rtn_name << "> (cid: " 
	    //	  outfile << "imp. stop ?: " << setw(1) << "<" << (IlpList.back())->name << "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	  tmpilp = (IlpProfile *)(IlpList.back());
	  tmpilp->dump(outfile);
	  tmpilp->dumpxml(outxml, call_depth+1);
	  outfile << endl;
	  call_depth++;
	}
      else if (count>0)
	{
	  {
	  //	  for (j=IlpList.size()-1; j>i; j--)
	  for (unsigned int j=0; j<count; j++ )
	    {
	      outfile << setw((call_depth+1)*5) << "";
	      outfile << "imp. stop: " << setw(1) << "<" << (IlpList.back())->name<< "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	      
	      tmpilp = (IlpProfile *)(IlpList.back());
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	      
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      (IlpList.back())->next_ilp = NULL;
	    }
	  //call_depth++;
	  }
	  // Perform the explicite STOP

	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpProfile *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }

	}
      else
	{
	  //      if (count)
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";// << " rtn_s_d: "<< rtn_start_depth << ")";// << endl; 
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpProfile *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }
	}
#ifdef DEBUG_ILP_VEC
  cerr << "----------DEBUG_ILP_VEC--CTRL_STOP--Start- " << rtn_image << "|" << rtn_name <<" ---- depth: "<< call_depth << "--- rtn_s_d: "<< rtn_start_depth << endl;

  for (unsigned int i=0; i<IlpList.size(); i++)
    cerr << "    IlpList[" << i << "] = "<< IlpList[i] << "   (" << IlpList[i]->name << ")"<< endl;
  cerr << endl;
  cerr << "----------DEBUG_ILP_VEC---CTRL_STOP-- Stop ------------" << endl;
#endif

    }


}

