
extern "C" {
#include "xed-interface.h"
#include "xed-immdis.h"
#include "xed-portability.h"
}

#include "pin.H"

#include "tools/ilptrace.hh"
#include "instrument.hh"

#include <iostream>
#include <sstream>
#include <iomanip>

#include "options.hh"
#include "extern.hh"

using std::cerr;
using std::endl;
using std::stringstream;


////////////////////////////////////////////////////////////////
//  CLASS ILPHISTOGRAM 
////////////////////////////////////////////////////////////////
IlpTrace::IlpTrace(const char *nickname, uint32_t dh, uint32_t cid):IlptoolInterface(nickname, dh, cid), trace()
{
      unsigned long i;
      for (i=0;i<REG_FILE_SIZE;i++){
	rat[i]=0;
	rat_trace[i]=0;
	nbread[i]=0;
	nbwrite[i]=0;
      }
      nbcycles=0; npg=0; numinst=0;
      nbcycles_control=0; numinst_control=0;

      memory.Reset();
      memory_trace.Reset();
      stoped = false;
      is_main = false;
      name = nickname;
      if(options.option_list["command"]=="trace")
	{
	  trace.resize((uint32_t)atoi(options.option_list["length"].c_str())+1);
	  gootrace.resize((uint32_t)atoi(options.option_list["length"].c_str())+1);
	}
}


IlpTrace::~IlpTrace() {}

///////////////////////////////////////////////////////////
////////// START OF /// ILP COMPUTE FUNCTIONS /////////////
///////////////////////////////////////////////////////////
VOID IlpTrace::analyse(Instruction &instr)

{
    uint32_t c1,cexe;
    uint32_t rdep[Instruction::MAXRRDEP] = {0, 0, 0, 0, 0, 0, 0, 0};
    uint32_t mdep[Instruction::MAXRRDEP] = {0, 0, 0, 0, 0, 0, 0, 0};
    uint32_t nbrdep;
    uint32_t nbmdep = 0, m, i, j, k=0,l,r;

    if(!stoped)
      {
	numinst++;

	//#define DEBUG_ANALYSE 1
#ifdef DEBUG_ANALYSE
	if (instr.nbr>Instruction::MAXRRDEP)
	  {
	    cerr << "ERROR instruction have too many sources !!!" << endl;
	    //	    cerr << instr << endl;
	    instr.dump(cerr);
	    cerr << endl;
	    abort();
	  }
#endif 
        for(i=0; i<instr.nbr; i++){
         for (j=0;j<instr.rr[i].len;j++){
          r=rat_trace[instr.rr[i].num+j];
          for(l=0;l<k;l++)
            if (r==rdep[l]) break;
          if (l==k){
            if (k<Instruction::MAXRRDEP){rdep[k]=r;k++;}
            else { cerr << "too many rdeps" << endl; abort();}
          }
         }
        }
        nbrdep=k;

        k=0;
        if (instr.amr){
          for(i=0; i<instr.rsize; i++){
            m=memory_trace.ReadMemory(instr.amr+i, 1);
            for(j=0; j<k; j++)
              if (m==mdep[j]) break;
            if (j==k){
              if (k<Instruction::MAXRRDEP) {mdep[k]=m;k++;}
              else { cerr << "too many mdeps" << endl; abort();}
            }
          }
        }
        if (instr.amr2){
          for(i=0; i<instr.rsize; i++){
            m=memory_trace.ReadMemory(instr.amr2+i, 1);
            for(j=0; j<k; j++)
              if (m==mdep[j]) break;
            if (j==k){
              if (k<Instruction::MAXRRDEP) {mdep[k]=m;k++;}
              else { cerr << "too many mdeps" << endl; abort();}
            }
          }
        }
        nbmdep=k;
	// CYCLE COMPUTATION
	stringstream tkey;
	tkey << hex << instr.address << dec;

	cexe = 0;
	for(i=0; i<nbrdep; i++)
	  {
	    c1 = trace[rdep[i]].get_cycle();
	    cexe = (c1>cexe)?c1:cexe;
	    //	    nbrdep++;
	    tkey << ":" << hex << trace[rdep[i]].get_address() << dec;
	  }
        for(i=0; i<nbmdep; i++){
	    c1 = trace[mdep[i]].get_cycle();
	    cexe = (c1>cexe)?c1:cexe;
	    tkey << ":" << hex << trace[mdep[i]].get_address() << dec;
	}
	cexe++;
	nbcycles=(cexe>nbcycles)?cexe:nbcycles;


	// Write to registers
	//#define DEBUG_ANALYSE 1
#ifdef DEBUG_ANALYSE
	if (instr.nbw>Instruction::MAXWRDEP)
	  {
	    cerr << "ERROR instruction have too many destinations !!!" << endl;
	    instr.dump(cerr);
	    cerr << endl;
	    abort();
	  }
#endif 
	// Get extra goo
	IlpTrace *current=this;
	while ( (current->next_ilp) != NULL && !(((IlpTrace *)(current->next_ilp))->stoped) )
	  {
	    current = (IlpTrace *)(current->next_ilp);
	  }
	
	GooOperation goooperation(instr.is_control, numinst, cexe, instr.address, instr.opcode, nbrdep, rdep, nbmdep, mdep, instr.xed_category, tkey.str().c_str(), (current->name).c_str(), current->call_id, current->call_depth);
	gootrace.update(goooperation, numinst);
	
	
	// TRACE UPDATE
	Operation operation(instr.is_control, numinst, cexe, instr.address, instr.opcode, nbrdep, rdep, nbmdep, mdep, instr.xed_category, tkey.str().c_str());

	if (nbmdep != operation.get_nbmdep() )
	  {
	    cerr << "DEBUG NBMDEP : nbmdep       = " << nbmdep << endl;
	    cerr << "DEBUG NBMDEP : get nbmdep   = " << operation.get_nbmdep() << endl;
	    
	    cerr << "DEBUG NBMDEP : get nbrdep   = " << operation.get_nbrdep() << endl;
	    cerr << endl;
	  }
	trace.update(operation, numinst);
        if (instr.is_control) trace.update_control(numinst,&numinst_control,&nbcycles_control);
	const char *opcode;
	char oponstack, opcontrol;
	opcode = OPCODE_StringShort(instr.opcode).c_str();
	oponstack = !(strcmp(opcode,"PUSH") &&
		      strcmp(opcode,"POP") &&
		      strcmp(opcode,"CALL_NEAR") &&
		      strcmp(opcode,"RET_NEAR"));
	opcontrol = (opcode[0]=='J') || !strcmp(opcode,"CALL_NEAR")
	  || !strcmp(opcode,"RET_NEAR")
	  || !strcmp(opcode,"SYSCALL");
	for(i=0; i<instr.nbw; i++)
	  {
#if defined(TARGET_IA32E)
	  if (oponstack && (instr.wr[i].num==RSP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#else
	  if (oponstack && (instr.wr[i].num==ESP)) continue;
	  if (opcontrol && (instr.wr[i].num==RIP)) continue;
#endif
	    for (j=0;j<instr.wr[i].len;j++){
	      rat_trace[instr.wr[i].num+j]=numinst;
	      rat[instr.wr[i].num+j]=cexe;
	    }
	  }
	
	if(instr.amw)
	  memory_trace.WriteMemory(numinst, instr.amw, instr.wsize);
      }
    if (next_ilp) next_ilp->analyse(instr);
  
}


///////////////////////////////////////////////////////////
////////// END OF /// ILP COMPUTE FUNCTIONS ///////////////
///////////////////////////////////////////////////////////

  // Pin calls this function every time a new instruction is encountered
VOID IlpTrace::instrument(INS ins, VOID *v)
    {
      UINT32 nr, nw, i;
      OPCODE op;
      Register rr[8], wr[8];
      IlpTrace *this_or_next = this;

      bool is_float=false;
/*
      if ( INS_RepPrefix(ins) || INS_RepnePrefix(ins) )
	{
	  // WARNING DO NOT USE REP Prefixed INSTRUCTION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
	  xed_decoded_inst_t* xedd = INS_XedDec(ins);
	  //	  xed_operand_values_t* xedv = INS_xed_operand_values(ins);
	  xed_operand_values_t* xedv = xed_decoded_inst_operands(xedd);
	  bool real_rep = xed_operand_values_has_real_rep(xedv);
	  if (real_rep)
	    {
	      // you have a 'real' rep prefix
	      cerr << "REP prefix detected and not supported yet !!!" << endl;
	      cerr << "Instruction address: " << hex << a << dec << endl;
	      cerr << INS_Disassemble(ins).c_str() << endl;
	      abort();
	    }
	}
*/

      op = INS_Opcode(ins);
      nr = INS_MaxNumRRegs(ins);

      if(nr>8) { cerr << "NR > 8 !!!" <<endl; abort(); }
      for (i=0; i<nr; i++){
	    rr[i].num = reg2int(INS_RegR(ins, i), &rr[i].len);
	    if (REG_is_fr(INS_RegR(ins, i))) is_float=true;
	    if (rr[i].num > 10000) { cerr << "RR > 10000 !!!" <<endl; abort(); }
      }
      nw = INS_MaxNumWRegs(ins);

      if(nw>8) { cerr << "NW > 8 !!!" <<endl; abort(); }
      for (i=0; i<nw; i++){
	wr[i].num = reg2int(INS_RegW(ins, i), &wr[i].len);
	if (REG_is_fr(INS_RegW(ins, i))) is_float=true;
	if (wr[i].num > 10000) 
	  { 
	    cerr << "WR > 10000 !!!: " << wr[i].num << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Mnemonic(ins) << endl;
	    cerr << hex << INS_Address(ins) << dec << "\t" << INS_Disassemble(ins) << endl;
	    abort();
	  }
	
      }

      UINT32 xed_category = INS_Category(ins);

     bool is_control;
     if (INS_IsBranchOrCall(ins)) is_control=true; else is_control=false;

     if ((options.option_list["floatonly"]=="0") || is_float)
	{

	  if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins) && INS_IsMemoryWrite(ins))
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	  else if (INS_IsMemoryRead(ins) && INS_HasMemoryRead2(ins))
	    INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_MEMORYREAD2_EA,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else if (INS_IsMemoryRead(ins) && INS_IsMemoryWrite(ins))
	  INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
			   IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else if (INS_IsMemoryRead(ins))
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_MEMORYREAD_EA,
			   IARG_MEMORYREAD_SIZE,
		           IARG_ADDRINT, 0L,
		           IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);


	else if (INS_IsMemoryWrite(ins))
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
		           IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
		           IARG_ADDRINT, 0L,
		           IARG_UINT32, 0,
		           IARG_ADDRINT, 0L,
			   IARG_MEMORYWRITE_EA,
			   IARG_MEMORYWRITE_SIZE,
			   IARG_UINT32, xed_category,
			   IARG_END);

	else 
	INS_InsertCall(ins, IPOINT_BEFORE, AFUNPTR(InstrumentAnalyse),
			   IARG_PTR, this_or_next,
			   IARG_INST_PTR,
                           IARG_BOOL, (BOOL)is_control,
			   IARG_UINT32, (UINT32)op,
			   IARG_UINT32, nr,
			   IARG_UINT32, nw,
			   IARG_UINT32, rr[0].num,
			   IARG_UINT32, rr[0].len,
			   IARG_UINT32, rr[1].num,
			   IARG_UINT32, rr[1].len,
			   IARG_UINT32, rr[2].num,
			   IARG_UINT32, rr[2].len,
			   IARG_UINT32, rr[3].num,
			   IARG_UINT32, rr[3].len,
			   IARG_UINT32, rr[4].num,
			   IARG_UINT32, rr[4].len,
			   IARG_UINT32, rr[5].num,
			   IARG_UINT32, rr[5].len,
			   IARG_UINT32, rr[6].num,
			   IARG_UINT32, rr[6].len,
			   IARG_UINT32, rr[7].num,
			   IARG_UINT32, rr[7].len,
			   IARG_UINT32, wr[0].num,
			   IARG_UINT32, wr[0].len,
			   IARG_UINT32, wr[1].num,
			   IARG_UINT32, wr[1].len,
			   IARG_UINT32, wr[2].num,
			   IARG_UINT32, wr[2].len,
			   IARG_UINT32, wr[3].num,
			   IARG_UINT32, wr[3].len,
			   IARG_UINT32, wr[4].num,
			   IARG_UINT32, wr[4].len,
			   IARG_UINT32, wr[5].num,
			   IARG_UINT32, wr[5].len,
			   IARG_UINT32, wr[6].num,
			   IARG_UINT32, wr[6].len,
			   IARG_UINT32, wr[7].num,
			   IARG_UINT32, wr[7].len,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_ADDRINT, 0L,
			   IARG_ADDRINT, 0L,
			   IARG_UINT32, 0,
			   IARG_UINT32, xed_category,
			   IARG_END);

	}
}

void IlpTrace::dump(ostream &os)
{
  os << "::I[" << numinst << "]";
  os << "::C[" << nbcycles << "]";
  os << "::ILP[" << (double)numinst/(double)nbcycles << "]" ;
  os << "::Ictrl[" << numinst_control << "]";
  os << "::Cctrl[" << nbcycles_control << "]";
  os << "::ILPctrl[" << (double)numinst_control/(double)nbcycles_control << "]" ;
  if(options.option_list["command"]=="trace")
    {
	  static int version = 0;
	  stringstream filename;
	  filename.imbue(std::locale("C"));
	  filename << options.option_list["resdir"] << "/";
	  filename << name << "_trace_depth_" << RoutineTrace::call_depth+1 << "_cid_" << call_id << "_ilptool.trace" ;
	  ofstream ofile(filename.str().c_str());
	  ofile.imbue(std::locale("C"));
	  trace.dump(ofile, numinst);
	  ofile.close();
	  stringstream filenamegoo;
	  filenamegoo.imbue(std::locale("C"));
	  filenamegoo << options.option_list["resdir"] << "/";
	  filenamegoo << name << "_trace_depth_" << RoutineTrace::call_depth+1 << "_cid_" << call_id << "_ilptool.gootrace" ;
	  version++;
	  ofstream ofilegoo(filenamegoo.str().c_str());
	  ofilegoo.imbue(std::locale("C"));
	  gootrace.dump(ofilegoo, numinst);
	  ofilegoo.close();
    }
}
void IlpTrace::dumpxml(ostream &os, int depth)
{
}

void IlpTrace::stop()
{
  stoped = true;
}

////////////////////////////////////////////////////////////////
//            END INTERFACE
////////////////////////////////////////////////////////////////

  

////////////////////////////////////////////////////////////////
//  CLASS ROUTINETRACE 
////////////////////////////////////////////////////////////////
RoutineTrace::RoutineTrace(const char *nname, const char *image, ADDRINT addr, RTN rtn, IlptoolInterface *ilp):
  RoutineInterface(nname, image, addr, rtn, 0, 0, 0, NULL)
{
  rtn_name = nname;
  if (ilp)
    call_vector.push_back(ilp);
}

RoutineTrace::~RoutineTrace() {}

void RoutineTrace::start()
{
  if (ctrl_start())
    {
      IlptoolInterface *new_ilp = new IlpTrace(this->rtn_name.c_str(), this->call_depth, this->call_id);
      (IlpList.back())->next_ilp = new_ilp;
      IlpList.push_back(new_ilp);
      call_vector.push_back(new_ilp);
      outfile << setw(call_depth*5) << "";
      outfile << "start : " << setw(15) << rtn_name << setw(10) << "   (depth: "<< call_depth << setw(10) << "rtn_s_d: "<< rtn_start_depth << ")" << endl;   
    }
}

void RoutineTrace::stop()
{
  IlpTrace *tmpilp=0;

  if (ctrl_stop())
    {

      // FIND FUNCTION which is STOP into IlpList vector.
      unsigned int i;
      unsigned int count=0;
      for (i=IlpList.size()-1; i>0; i--)
	{
	  if (rtn_name == IlpList[i]->name)
	    break;
	  count++;
	}
      
      if (count == IlpList.size()-1 )
	{
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "imp. stop ?: " << setw(1) << "<" << rtn_name << "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	  tmpilp = (IlpTrace *)(IlpList.back());
	  tmpilp->dump(outfile);
	  tmpilp->dumpxml(outxml, call_depth+1);
	  outfile << endl;
	  call_depth++;
	}
      else if (count>0)
	{
	  {
	  for (unsigned int j=0; j<count; j++ )
	    {
	      outfile << setw((call_depth+1)*5) << "";
	      outfile << "imp. stop: " << setw(1) << "<" << (IlpList.back())->name<< "> (cid: " 
		  << (IlpList.back())->call_id << ") "
		  << "> (lcid: " 
		  << local_cid << ") ";
	      
	      tmpilp = (IlpTrace *)(IlpList.back());
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	      
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      (IlpList.back())->next_ilp = NULL;
	    }
	  }
	  // Perform the explicite STOP

	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";// << " rtn_s_d: "<< rtn_start_depth << ")";// << endl; 
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpTrace *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }

	}
      else
	{
	  //      if (count)
	  outfile << setw((call_depth+1)*5) << "";
	  outfile << "stop  : " << setw(1) << "<" << rtn_name << ">" ;
	  outfile << "   (depth: "<< (call_depth+1) << ", lcid: "<< local_cid <<")";// << " rtn_s_d: "<< rtn_start_depth << ")";// << endl; 
	  
	  // XML Output
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<name>"<< rtn_name << "</name>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<cid>"<< (IlpList.back())->call_id << "</cid>"<<endl;
	  
	  outxml << setw((call_depth+2)*5) << "";
	  outxml << "<depth>"<< call_depth+1 << "</depth>"<<endl;
	  

	  if (IlpList.size()>1)
	    {
	      tmpilp = (IlpTrace *)IlpList.back();
	      tmpilp->dump(outfile);
	      tmpilp->dumpxml(outxml, call_depth+1);
	      outfile << endl;
	    }
	  
	  outxml << setw((call_depth+1)*5) << "";
	  outxml << "</call>" << endl;
	  
	  if(IlpList.size()>1)
	    {
	      (IlpList.back())->stop();
	      IlpList.pop_back();
	      delete tmpilp;
	      if (IlpList.size()>0)
		(IlpList.back())->next_ilp = NULL;
	    }
	}
#ifdef DEBUG_ILP_VEC
      cerr << "----------DEBUG_ILP_VEC--CTRL_STOP--Start- " << rtn_image << "|" << rtn_name <<" ---- depth: "<< call_depth << "--- rtn_s_d: "<< rtn_start_depth << endl;
      for (unsigned int i=0; i<IlpList.size(); i++)
	cerr << "    IlpList[" << i << "] = "<< IlpList[i] << "   (" << IlpList[i]->name << ")"<< endl;
      
      cerr << endl;
      cerr << "----------DEBUG_ILP_VEC---CTRL_STOP-- Stop ------------" << endl;
#endif

    }
  
}

